# Copyright (C) 2008 Robert C. Kirby (Texas Tech University)
# Modified by Andrew T. T. McRae (Imperial College London)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import itertools
import math
import numpy as np
from FIAT import finite_element, polynomial_set, dual_set, functional, P0
from FIAT.reference_element import LINE, make_lattice
from FIAT.orientation_utils import make_entity_permutations_simplex
from FIAT.barycentric_interpolation import LagrangePolynomialSet, get_lagrange_points
from FIAT.polynomial_set import mis
from FIAT.check_format_variant import parse_lagrange_variant


def make_entity_permutations(dim, npoints):
    if npoints <= 0:
        return {o: [] for o in range(math.factorial(dim + 1))}
    # DG nodes are numbered, in order of significance,
    # - by g0: entity dim (vertices first, then edges, then ...)
    # - by g1: entity ids (DoFs on entities of smaller ids first)
    # - lexicographically as in CG
    #
    # Example:
    # dim = 2, degree = 3 (npoints = degree + 1)
    #
    #     facet ids     Lexicographic  DG (degree = 3)
    #    +              DoF numbering  DoF numbering
    #    | \            9              2
    #    |   \  0       7 8            6 4
    #  1 |     \        4 5 6          5 9 3
    #    |       \      0 1 2 3        0 7 8 1
    #    +--------+
    #        2
    # where DG degrees of freedom are numbered to geometrically
    # coincide with CG degrees of freedom.
    #
    # In the below we will show example outputs corresponding to
    # the above example.

    a = np.array(sorted(mis(dim + 1, npoints - 1)), dtype=int)
    # >>> a
    # [[0, 0, 3],
    #  [0, 1, 2],   # (3,0,0)
    #  [0, 2, 1],   #
    #  [0, 3, 0],   # (2,0,1)(2,1,0)
    #  [1, 0, 2],   #
    #  [1, 1, 1],   # (1,0,2)(1,1,1)(1,2,0)
    #  [1, 2, 0],   #
    #  [2, 0, 1],   # (0,0,3)(0,1,2)(0,2,1)(0,3,0)
    #  [2, 1, 0],   #
    #  [3, 0, 0]]   # Lattice points that a represents
    # a.shape[0] = number of DoFs
    # a.shape[1] = dim + 1
    # sum of each row = degree (bary centric lattice coordinates)

    # Flip along the axis 1 for convenience.
    # This would make: np.lexsort(a.transpose()) = [0, 1, 2, ..., 9].
    a = a[:, ::-1]

    index_perms = sorted(itertools.permutations(range(dim + 1)))
    # >>> index_perms
    # [[0, 1, 2],
    #  [0, 2, 1],
    #  [1, 0, 2],
    #  [1, 2, 0],
    #  [2, 0, 1],
    #  [2, 1, 0]]

    # First separate by entity dimension.
    g0 = dim - (a == 0).astype(int).sum(axis=1)
    # >>> g0
    # [ 0, 1, 1, 0, 1, 2, 1, 1, 1, 0]  # 0 for vertices
    #                                  # 1 for edges
    #                                  # 2 for cell

    # Then separate by entity number.
    g1 = np.zeros_like(g0)
    for d in range(dim + 1):
        on_facet_d = (a[:, d] == 0).astype(int)
        g1 += d * on_facet_d
    # The above logic is consistent with the FIAT entity numbering
    # convention ("entities that are not incident to vertex 0 are
    # numbered first, then ..."), but vertices are not numbered using
    # the same logic in FIAT, so we need to fix:
    g1[g0 == 0] = -g1[g0 == 0]
    # >>> g1
    # [-3, 2, 2,-2, 1, 0, 0, 1, 0,-1]

    # Compoute the map from the DG DoFs to the lattice points on the cell.
    # For each entity dimension, DoFs that have smaller numbers in g1
    # will be assigned smaller DG node numbers.
    # Order first by g0, then by g1, and finally by a (lexicographically as in CG)
    g0 = g0.reshape((a.shape[0], 1))
    g1 = g1.reshape((a.shape[0], 1))
    dg_to_lattice = np.lexsort(np.transpose(np.concatenate((a, g1, g0), axis=1)))
    # >>> dg_to_lattice
    # [ 0, 3, 9, 6, 8, 4, 7, 1, 2, 5]

    # Compute the inverse map.
    lattice_to_dg = np.empty_like(dg_to_lattice)
    for i, im in enumerate(dg_to_lattice):
        lattice_to_dg[im] = i
    # >>> lattice_to_dg
    # [ 0, 7, 8, 1, 5, 9, 3, 6, 4, 2]
    perms = {}
    for o, index_perm in enumerate(index_perms):
        # First compute permutation in lattice point numbers in lattice point order (as we do for CG cell DoFs)
        perm = np.lexsort(np.transpose(a[:, index_perm]))
        # Then convert to DG DoF numbers in DG DoF order:
        # lattice_to_dg[perm]                -> convert lattice point numbers to DG DoF numbers
        # lattice_to_dg[perm][dg_to_lattice] -> reorder for DG DoF order
        #
        # Example:
        # Under one CW rotation, a DG element on a physical cell
        # is mapped to the FIAT reference as:
        #
        # 0
        # 7 5
        # 8 9 6
        # 1 3 4 2
        #
        # Under the same transformation, the lattice points would
        # be mapped as:
        #
        # 0
        # 1 4
        # 2 5 7
        # 3 6 8 9
        #
        # In this case we will have:
        #
        # perm                               = [3, 6, 8, 9, 2, 5, 7, 1, 4, 0]
        # lattice_to_dg[perm]                = [1, 3, 4, 2, 8, 9, 6, 7, 5, 0]
        # lattice_to_dg[perm][dg_to_lattice] = [1, 2, 0, 6, 5, 8, 7, 3, 4, 9]
        #
        # Note:
        # Sane thing to do is to just number DG dofs on a lattice.
        perm = lattice_to_dg[perm][dg_to_lattice]
        perms[o] = perm.tolist()
    return perms


class BrokenLagrangeDualSet(dual_set.DualSet):
    """The dual basis for Lagrange elements.  This class works for
    simplices of any dimension.  Nodes are point evaluation at
    equispaced points.  This is the broken version where
    all nodes are topologically associated with the cell itself"""

    def __init__(self, ref_el, degree, point_variant="equispaced"):
        nodes = []
        entity_ids = {}
        entity_permutations = {}

        # make nodes by getting points
        # need to do this dimension-by-dimension, facet-by-facet
        top = ref_el.get_topology()
        for dim in sorted(top):
            entity_ids[dim] = {}
            entity_permutations[dim] = {}
            perms = make_entity_permutations(dim, degree + 1 if dim == len(top) - 1 else -1)
            for entity in sorted(top[dim]):
                pts_cur = ref_el.make_points(dim, entity, degree, variant=point_variant)
                nodes.extend(functional.PointEvaluation(ref_el, x) for x in pts_cur)
                entity_ids[dim][entity] = []
                entity_permutations[dim][entity] = perms
        entity_ids[dim][0] = list(range(len(nodes)))

        super().__init__(nodes, ref_el, entity_ids, entity_permutations)


class DiscontinuousLagrangeDualSet(dual_set.DualSet):
    """The dual basis for discontinuous elements with nodes at recursively-defined points.
    """
    def __init__(self, ref_el, degree, point_variant="equispaced"):
        nodes = []
        entity_ids = {}
        entity_permutations = {}
        sd = ref_el.get_dimension()
        top = ref_el.get_topology()
        for dim in sorted(top):
            entity_ids[dim] = {}
            entity_permutations[dim] = {}
            perms = make_entity_permutations_simplex(dim, degree + 1 if dim == sd else -1)
            for entity in sorted(top[dim]):
                entity_ids[dim][entity] = []
                entity_permutations[dim][entity] = perms

        # make nodes by getting points on the interior facets
        for entity in top[sd]:
            cur = len(nodes)
            pts = make_lattice(ref_el.get_vertices_of_subcomplex(top[sd][entity]),
                               degree, variant=point_variant)
            nodes.extend(functional.PointEvaluation(ref_el, x) for x in pts)
            entity_ids[dim][entity] = list(range(cur, len(nodes)))
        super().__init__(nodes, ref_el, entity_ids, entity_permutations)


class DiscontinuousLagrange(finite_element.CiarletElement):
    """The discontinuous Lagrange finite element.

    :arg ref_el:  The reference element, which could be a standard FIAT simplex or a split complex
    :arg degree:  The polynomial degree
    :arg variant: A comma-separated string that may specify the type of point distribution
                  and the splitting strategy if a macro element is desired.
                  Either option may be omitted.  The default point type is equispaced
                  and the default splitting strategy is None.
                  Example: variant='gl' gives a standard unsplit point distribution with
                              spectral points.
                           variant='equispaced,Iso(2)' with degree=1 gives the P2:P1 iso element.
                           variant='Alfeld' can be used to obtain a barycentrically refined
                              macroelement for Scott-Vogelius.
    """
    def __new__(cls, ref_el, degree, variant="equispaced"):
        if degree == 0:
            splitting, _ = parse_lagrange_variant(variant, discontinuous=True)
            if splitting is None and not ref_el.is_macrocell():
                # FIXME P0 on the split requires implementing SplitSimplicialComplex.symmetry_group_size()
                return P0.P0(ref_el)
        return super().__new__(cls)

    def __init__(self, ref_el, degree, variant="equispaced"):
        splitting, point_variant = parse_lagrange_variant(variant, discontinuous=True)
        if splitting is not None:
            ref_el = splitting(ref_el)
        if point_variant in ("equispaced", "gll", "lgc"):
            dual = BrokenLagrangeDualSet(ref_el, degree, point_variant=point_variant)
        else:
            dual = DiscontinuousLagrangeDualSet(ref_el, degree, point_variant=point_variant)
        if ref_el.shape == LINE:
            # In 1D we can use the primal basis as the expansion set,
            # avoiding any round-off coming from a basis transformation
            points = get_lagrange_points(dual)
            poly_set = LagrangePolynomialSet(ref_el, points)
        else:
            poly_set = polynomial_set.ONPolynomialSet(ref_el, degree)
        formdegree = ref_el.get_spatial_dimension()  # n-form
        super().__init__(poly_set, dual, degree, formdegree)



import pytest
import numpy

from FIAT import (GopalakrishnanLedererSchoberlFirstKind,
                  GopalakrishnanLedererSchoberlSecondKind)
from FIAT.reference_element import ufc_simplex
from FIAT.expansions import polynomial_dimension
from FIAT.polynomial_set import ONPolynomialSet
from FIAT.quadrature_schemes import create_quadrature
from FIAT.quadrature import FacetQuadratureRule


@pytest.fixture(params=("T", "S"))
def cell(request):
    dim = {"I": 1, "T": 2, "S": 3}[request.param]
    return ufc_simplex(dim)


@pytest.mark.parametrize("degree", (1, 2, 3))
@pytest.mark.parametrize("kind", (1, 2))
def test_gls_bubbles(kind, cell, degree):
    if kind == 1:
        element = GopalakrishnanLedererSchoberlFirstKind
    else:
        element = GopalakrishnanLedererSchoberlSecondKind
    fe = element(cell, degree)
    sd = cell.get_spatial_dimension()
    facet_el = cell.construct_subelement(sd-1)
    poly_set = fe.get_nodal_basis()

    # test dimension of constrained space
    dimPkm1 = polynomial_dimension(facet_el, degree-1)
    dimPkp1 = polynomial_dimension(facet_el, degree+1)
    dimPk = polynomial_dimension(facet_el, degree)
    if kind == 1:
        constraints = dimPk - dimPkm1
    else:
        constraints = 0
    expected = (sd**2-1)*(polynomial_dimension(cell, degree) - constraints)
    assert poly_set.get_num_members() == expected

    # test dimension of the bubbles
    entity_dofs = fe.entity_dofs()
    bubbles = poly_set.take(entity_dofs[sd][0])
    expected = (sd**2-1)*polynomial_dimension(cell, degree-1)
    assert bubbles.get_num_members() == expected

    top = cell.get_topology()
    Qref = create_quadrature(facet_el, 2*degree+1)
    Pk = ONPolynomialSet(facet_el, degree+1)
    if kind == 1:
        start, stop = dimPkm1, dimPkp1
    else:
        start, stop = dimPk, dimPkp1
    PkH = Pk.take(list(range(start, stop)))
    PkH_at_qpts = PkH.tabulate(Qref.get_points())[(0,)*(sd-1)]
    weights = numpy.transpose(numpy.multiply(PkH_at_qpts, Qref.get_weights()))
    for facet in top[sd-1]:
        n = cell.compute_scaled_normal(facet)
        rts = cell.compute_tangents(sd-1, facet)
        Q = FacetQuadratureRule(cell, sd-1, facet, Qref)
        qpts, qwts = Q.get_points(), Q.get_weights()

        # test the degree of normal-tangential components
        phi_at_pts = fe.tabulate(0, qpts)[(0,) * sd]
        for t in rts:
            nt = numpy.outer(t, n)
            phi_nt = numpy.tensordot(nt, phi_at_pts, axes=((0, 1), (1, 2)))
            assert numpy.allclose(numpy.dot(phi_nt, weights), 0)

        # test the support of the normal-tangential bubble
        phi_at_pts = bubbles.tabulate(qpts)[(0,) * sd]
        for t in rts:
            nt = numpy.outer(t, n)
            phi_nt = numpy.tensordot(nt, phi_at_pts, axes=((0, 1), (1, 2)))
            norms = numpy.dot(phi_nt**2, qwts)
            assert numpy.allclose(norms, 0)



from FIAT.hdivcurl import Hdiv, Hcurl
from FIAT.reference_element import LINE

import gem
from gem.utils import cached_property
from finat.finiteelementbase import FiniteElementBase
from finat.tensor_product import TensorProductElement


class WrapperElementBase(FiniteElementBase):
    """Common base class for H(div) and H(curl) element wrappers."""

    def __init__(self, wrappee, transform):
        super().__init__()
        self.wrappee = wrappee
        """An appropriate tensor product FInAT element whose basis
        functions are mapped to produce an H(div) or H(curl)
        conforming element."""

        self.transform = transform
        """A transformation applied on the scalar/vector values of the
        wrapped element to produce an H(div) or H(curl) conforming
        element."""

    @property
    def cell(self):
        return self.wrappee.cell

    @property
    def complex(self):
        return self.wrappee.complex

    @property
    def degree(self):
        return self.wrappee.degree

    def entity_dofs(self):
        return self.wrappee.entity_dofs()

    @property
    def entity_permutations(self):
        return self.wrappee.entity_permutations

    def entity_closure_dofs(self):
        return self.wrappee.entity_closure_dofs()

    def entity_support_dofs(self):
        return self.wrappee.entity_support_dofs()

    def space_dimension(self):
        return self.wrappee.space_dimension()

    @property
    def index_shape(self):
        return self.wrappee.index_shape

    @property
    def value_shape(self):
        return (self.cell.get_spatial_dimension(),)

    def _transform_evaluation(self, core_eval):
        beta = self.get_indices()
        zeta = self.get_value_indices()

        def promote(table):
            v = gem.partial_indexed(table, beta)
            u = gem.ListTensor(self.transform(v))
            return gem.ComponentTensor(gem.Indexed(u, zeta), beta + zeta)

        return {alpha: promote(table)
                for alpha, table in core_eval.items()}

    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        core_eval = self.wrappee.basis_evaluation(order, ps, entity)
        return self._transform_evaluation(core_eval)

    def point_evaluation(self, order, refcoords, entity=None, coordinate_mapping=None):
        core_eval = self.wrappee.point_evaluation(order, refcoords, entity)
        return self._transform_evaluation(core_eval)

    @property
    def dual_basis(self):
        Q, x = self.wrappee.dual_basis
        beta = self.get_indices()
        zeta = self.get_value_indices()
        # Index out the basis indices from wrapee's Q, to get
        # something of wrappee.value_shape, then promote to new shape
        # with the same transform as done for basis evaluation
        Q = gem.ListTensor(self.transform(gem.partial_indexed(Q, beta)))
        # Finally wrap up Q in shape again (now with some extra
        # value_shape indices)
        return gem.ComponentTensor(Q[zeta], beta + zeta), x


class HDivElement(WrapperElementBase):
    """H(div) wrapper element for tensor product elements."""

    def __init__(self, wrappee):
        assert isinstance(wrappee, TensorProductElement)
        if any(fe.formdegree is None for fe in wrappee.factors):
            raise ValueError("Form degree of subelement is None, cannot H(div)!")

        formdegree = sum(fe.formdegree for fe in wrappee.factors)
        if formdegree != wrappee.cell.get_spatial_dimension() - 1:
            raise ValueError("H(div) requires (n-1)-form element!")

        transform = select_hdiv_transformer(wrappee)
        super().__init__(wrappee, transform)

    @property
    def formdegree(self):
        return self.cell.get_spatial_dimension() - 1

    @cached_property
    def fiat_equivalent(self):
        return Hdiv(self.wrappee.fiat_equivalent)

    @property
    def mapping(self):
        return "contravariant piola"


class HCurlElement(WrapperElementBase):
    """H(curl) wrapper element for tensor product elements."""

    def __init__(self, wrappee):
        assert isinstance(wrappee, TensorProductElement)
        if any(fe.formdegree is None for fe in wrappee.factors):
            raise ValueError("Form degree of subelement is None, cannot H(curl)!")

        formdegree = sum(fe.formdegree for fe in wrappee.factors)
        if formdegree != 1:
            raise ValueError("H(curl) requires 1-form element!")

        transform = select_hcurl_transformer(wrappee)
        super().__init__(wrappee, transform)

    @property
    def formdegree(self):
        return 1

    @cached_property
    def fiat_equivalent(self):
        return Hcurl(self.wrappee.fiat_equivalent)

    @property
    def mapping(self):
        return "covariant piola"


def select_hdiv_transformer(element):
    # Assume: something x interval
    assert len(element.factors) == 2
    assert element.factors[1].cell.get_shape() == LINE

    # Globally consistent edge orientations of the reference
    # quadrilateral: rightward horizontally, upward vertically.
    # Their rotation by 90 degrees anticlockwise is interpreted as the
    # positive direction for normal vectors.
    ks = tuple(fe.formdegree for fe in element.factors)
    if ks == (0, 1):
        # Make the scalar value the leftward-pointing normal on the
        # y-aligned edges.
        return lambda v: [gem.Product(gem.Literal(-1), v), gem.Zero()]
    elif ks == (1, 0):
        # Make the scalar value the upward-pointing normal on the
        # x-aligned edges.
        return lambda v: [gem.Zero(), v]
    elif ks == (2, 0):
        # Same for 3D, so z-plane.
        return lambda v: [gem.Zero(), gem.Zero(), v]
    elif ks == (1, 1):
        if element.mapping == "contravariant piola":
            # Pad the 2-vector normal on the "base" cell into a
            # 3-vector, maintaining direction.
            return lambda v: [gem.Indexed(v, (0,)),
                              gem.Indexed(v, (1,)),
                              gem.Zero()]
        elif element.mapping == "covariant piola":
            # Rotate the 2-vector tangential component on the "base"
            # cell 90 degrees anticlockwise into a 3-vector and pad.
            return lambda v: [gem.Indexed(v, (1,)),
                              gem.Product(gem.Literal(-1), gem.Indexed(v, (0,))),
                              gem.Zero()]
        else:
            assert False, "Unexpected original mapping!"
    else:
        assert False, "Unexpected form degree combination!"


def select_hcurl_transformer(element):
    # Assume: something x interval
    assert len(element.factors) == 2
    assert element.factors[1].cell.get_shape() == LINE

    # Globally consistent edge orientations of the reference
    # quadrilateral: rightward horizontally, upward vertically.
    # Tangential vectors interpret these as the positive direction.
    dim = element.cell.get_spatial_dimension()
    ks = tuple(fe.formdegree for fe in element.factors)
    if element.mapping == "affine":
        if ks == (1, 0):
            # Can only be 2D.  Make the scalar value the
            # rightward-pointing tangential on the x-aligned edges.
            return lambda v: [v, gem.Zero()]
        elif ks == (0, 1):
            # Can be any spatial dimension.  Make the scalar value the
            # upward-pointing tangential.
            return lambda v: [gem.Zero()] * (dim - 1) + [v]
        else:
            assert False
    elif element.mapping == "covariant piola":
        # Second factor must be continuous interval.  Just padding.
        return lambda v: [gem.Indexed(v, (0,)),
                          gem.Indexed(v, (1,)),
                          gem.Zero()]
    elif element.mapping == "contravariant piola":
        # Second factor must be continuous interval.  Rotate the
        # 2-vector tangential component on the "base" cell 90 degrees
        # clockwise into a 3-vector and pad.
        return lambda v: [gem.Product(gem.Literal(-1), gem.Indexed(v, (1,))),
                          gem.Indexed(v, (0,)),
                          gem.Zero()]
    else:
        assert False, "Unexpected original mapping!"



import FIAT

from finat.physically_mapped import Citations
from finat.piola_mapped import PiolaBubbleElement


class GuzmanNeilanFirstKindH1(PiolaBubbleElement):
    """Pk^d enriched with Guzman-Neilan bubbles."""
    def __init__(self, cell, order=1):
        if Citations is not None:
            Citations().register("GuzmanNeilan2018")
        super().__init__(FIAT.GuzmanNeilanFirstKindH1(cell, order=order))


class GuzmanNeilanSecondKindH1(PiolaBubbleElement):
    """C0 Pk^d(Alfeld) enriched with Guzman-Neilan bubbles."""
    def __init__(self, cell, order=1):
        if Citations is not None:
            Citations().register("GuzmanNeilan2018")
        super().__init__(FIAT.GuzmanNeilanSecondKindH1(cell, order=order))


class GuzmanNeilanBubble(GuzmanNeilanFirstKindH1):
    """Modified Bernardi-Raugel bubbles that are C^0 P_dim(Alfeld) with constant divergence."""
    def __init__(self, cell, degree=None):
        super().__init__(cell, order=0)


class GuzmanNeilanH1div(PiolaBubbleElement):
    """Alfeld-Sorokina nodally enriched with Guzman-Neilan bubbles."""
    def __init__(self, cell, degree=None):
        if Citations is not None:
            Citations().register("GuzmanNeilan2018")
        super().__init__(FIAT.GuzmanNeilanH1div(cell, degree=degree))



# Copyright (C) 2008-2012 Robert C. Kirby (Texas Tech University)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

# polynomial sets
# basic interface:
# -- defined over some reference element
# -- need to be able to tabulate (jets)
# -- type of entry: could by scalar, numpy array, or object-value
#    (such as symmetric tensors, as long as they can be converted <-->
#    with 1d arrays)
# Don't need the "Polynomial" class we had before, provided that
# we have an interface for defining sets of functionals (moments against
# an entire set of polynomials)

import numpy
from itertools import chain
from FIAT import expansions
from FIAT.functional import index_iterator


def mis(m, n):
    """Returns all m-tuples of nonnegative integers that sum up to n."""
    if m == 1:
        return [(n,)]
    elif n == 0:
        return [(0,) * m]
    else:
        return [(n - i,) + foo
                for i in range(n + 1)
                for foo in mis(m - 1, i)]


# We order coeffs by C_{i,j,k}
# where i is the index into the polynomial set,
# j may be an empty tuple (scalar polynomials)
#   or else a vector/tensor
# k is the expansion function
# so if I have all bfs at a given point x in an array bf,
# then dot(coeffs, bf) gives the array of bfs
class PolynomialSet(object):
    """Implements a set of polynomials as linear combinations of an
    expansion set over a reference element.
    ref_el: the reference element
    degree: an order labeling the space
    embedded degree: the degree of polynomial expansion basis that
         must be used to evaluate this space
    coeffs: A numpy array containing the coefficients of the expansion
         basis for each member of the set.  Coeffs is ordered by
         coeffs[i,j,k] where i is the label of the member, k is
         the label of the expansion function, and j is a (possibly
         empty) tuple giving the index for a vector- or tensor-valued
         function.
    """
    def __init__(self, ref_el, degree, embedded_degree, expansion_set, coeffs):
        self.ref_el = ref_el
        self.num_members = coeffs.shape[0]
        self.degree = degree
        self.embedded_degree = embedded_degree
        self.expansion_set = expansion_set
        self.coeffs = coeffs

    def tabulate_new(self, pts):
        return numpy.dot(self.coeffs,
                         self.expansion_set.tabulate(self.embedded_degree, pts))

    def tabulate(self, pts, jet_order=0):
        """Returns the values of the polynomial set."""
        base_vals = self.expansion_set._tabulate(self.embedded_degree, pts, order=jet_order)
        result = {alpha: numpy.dot(self.coeffs, base_vals[alpha]) for alpha in base_vals}
        return result

    def get_expansion_set(self):
        return self.expansion_set

    def get_coeffs(self):
        return self.coeffs

    def get_num_members(self):
        return self.num_members

    def get_degree(self):
        return self.degree

    def get_embedded_degree(self):
        return self.embedded_degree

    def get_dmats(self, cell=0):
        return self.expansion_set.get_dmats(self.embedded_degree, cell=cell)

    def get_reference_element(self):
        return self.ref_el

    def get_shape(self):
        """Returns the shape of phi(x), where () corresponds to
        scalar (2,) a vector of length 2, etc"""
        return self.coeffs.shape[1:-1]

    def take(self, items):
        """Extracts subset of polynomials given by items."""
        new_coeffs = numpy.take(self.get_coeffs(), items, 0)
        return PolynomialSet(self.ref_el, self.degree, self.embedded_degree,
                             self.expansion_set, new_coeffs)

    def __len__(self):
        return self.num_members


class ONPolynomialSet(PolynomialSet):
    """Constructs an orthonormal basis out of expansion set by having an
    identity matrix of coefficients.  Can be used to specify ON bases
    for vector- and tensor-valued sets as well.
    """
    def __init__(self, ref_el, degree, shape=tuple(), **kwargs):
        expansion_set = expansions.ExpansionSet(ref_el, **kwargs)
        if shape == tuple():
            num_components = 1
        else:
            flat_shape = numpy.ravel(shape)
            num_components = numpy.prod(flat_shape)
        num_exp_functions = expansion_set.get_num_members(degree)
        num_members = num_components * num_exp_functions
        embedded_degree = degree

        # set up coefficients
        if shape == tuple():
            coeffs = numpy.eye(num_members)
        else:
            coeffs = numpy.zeros((num_members, *shape, num_exp_functions))
            cur = 0
            exp_bf = range(num_exp_functions)
            for idx in index_iterator(shape):
                cur_bf = range(cur, cur+num_exp_functions)
                coeffs[(cur_bf, *idx, exp_bf)] = 1.0
                cur += num_exp_functions

        super().__init__(ref_el, degree, embedded_degree, expansion_set, coeffs)


def project(f, U, Q):
    """Computes the expansion coefficients of f in terms of the members of
    a polynomial set U.  Numerical integration is performed by
    quadrature rule Q.
    """
    pts = Q.get_points()
    wts = Q.get_weights()
    f_at_qps = [f(x) for x in pts]
    U_at_qps = U.tabulate(pts)
    coeffs = numpy.array([sum(wts * f_at_qps * phi) for phi in U_at_qps])
    return coeffs


def form_matrix_product(mats, alpha):
    """Forms product over mats[i]**alpha[i]"""
    m = mats[0].shape[0]
    result = numpy.eye(m)
    for i in range(len(alpha)):
        for j in range(alpha[i]):
            result = numpy.dot(mats[i], result)
    return result


def spanning_basis(A, nullspace=False, rtol=1e-10):
    """Construct a basis that spans the rows of A via SVD.
    """
    Aflat = A.reshape(A.shape[0], -1)
    u, sig, vt = numpy.linalg.svd(Aflat, full_matrices=True)
    atol = rtol * (sig[0] + 1)
    num_sv = len([s for s in sig if abs(s) > atol])
    basis = vt[num_sv:] if nullspace else vt[:num_sv]
    return numpy.reshape(basis, (-1, *A.shape[1:]))


def polynomial_set_union_normalized(A, B):
    """Given polynomial sets A and B, constructs a new polynomial set
    whose span is the same as that of span(A) union span(B).  It may
    not contain any of the same members of the set, as we construct a
    span via SVD.
    """
    assert A.get_reference_element() == B.get_reference_element()
    new_coeffs = construct_new_coeffs(A.get_reference_element(), A, B)

    deg = max(A.get_degree(), B.get_degree())
    em_deg = max(A.get_embedded_degree(), B.get_embedded_degree())
    coeffs = spanning_basis(new_coeffs)
    return PolynomialSet(A.get_reference_element(),
                         deg,
                         em_deg,
                         A.get_expansion_set(),
                         coeffs)


def construct_new_coeffs(ref_el, A, B):
    """
    Constructs new coefficients for the set union of A and B
    If A and B are discontinuous and do not have the same degree the smaller one
    is extended to match the larger.

    This does not handle the case that A and B have continuity but not the same degree.
    """

    if A.get_expansion_set().continuity != B.get_expansion_set().continuity:
        raise ValueError("Continuity of expansion sets does not match.")

    if A.get_embedded_degree() != B.get_embedded_degree() and A.get_expansion_set().continuity is None:
        higher = A if A.get_embedded_degree() > B.get_embedded_degree() else B
        lower = B if A.get_embedded_degree() > B.get_embedded_degree() else A

        diff = higher.coeffs.shape[-1] - lower.coeffs.shape[-1]

        # pad only the 0th axis with the difference in size
        padding = [(0, 0) for i in range(len(lower.coeffs.shape) - 1)] + [(0, diff)]
        embedded_coeffs = numpy.pad(lower.coeffs, padding)

        new_coeffs = numpy.concatenate((embedded_coeffs, higher.coeffs), axis=0)
    elif A.get_embedded_degree() == B.get_embedded_degree():
        new_coeffs = numpy.concatenate((A.coeffs, B.coeffs), axis=0)
    else:
        raise NotImplementedError("Extending of coefficients is not implemented for PolynomialSets with continuity and different degrees")
    return new_coeffs


class ONSymTensorPolynomialSet(PolynomialSet):
    """Constructs an orthonormal basis for symmetric-tensor-valued
    polynomials on a reference element.
    """
    def __init__(self, ref_el, degree, size=None, **kwargs):
        expansion_set = expansions.ExpansionSet(ref_el, **kwargs)

        sd = ref_el.get_spatial_dimension()
        if size is None:
            size = sd

        shape = (size, size)
        num_exp_functions = expansion_set.get_num_members(degree)
        num_components = size * (size + 1) // 2
        num_members = num_components * num_exp_functions
        embedded_degree = degree

        # set up coefficients for symmetric tensors
        coeffs = numpy.zeros((num_members, *shape, num_exp_functions))
        cur = 0
        exp_bf = range(num_exp_functions)
        for i, j in index_iterator(shape):
            if i > j:
                continue
            cur_bf = range(cur, cur+num_exp_functions)
            coeffs[cur_bf, i, j, exp_bf] = 1.0
            coeffs[cur_bf, j, i, exp_bf] = 1.0
            cur += num_exp_functions

        super().__init__(ref_el, degree, embedded_degree, expansion_set, coeffs)


class TracelessTensorPolynomialSet(PolynomialSet):
    """Constructs an orthonormal basis for traceless-tensor-valued
    polynomials on a reference element.
    """
    def __init__(self, ref_el, degree, size=None, **kwargs):
        expansion_set = expansions.ExpansionSet(ref_el, **kwargs)

        sd = ref_el.get_spatial_dimension()
        if size is None:
            size = sd

        shape = (size, size)
        num_exp_functions = expansion_set.get_num_members(degree)
        num_components = size * size - 1
        num_members = num_components * num_exp_functions
        embedded_degree = degree

        # set up coefficients for traceless tensors
        coeffs = numpy.zeros((num_members, *shape, num_exp_functions))
        cur = 0
        exp_bf = range(num_exp_functions)
        for i, j in index_iterator(shape):
            if i == size-1 and j == size-1:
                continue
            cur_bf = range(cur, cur+num_exp_functions)
            coeffs[cur_bf, i, j, exp_bf] = 1.0
            if i == j:
                coeffs[cur_bf, -1, -1, exp_bf] = -1.0
            cur += num_exp_functions

        super().__init__(ref_el, degree, embedded_degree, expansion_set, coeffs)


def make_bubbles(ref_el, degree, codim=0, shape=(), scale="L2 piola"):
    """Construct a polynomial set with codim bubbles up to the given degree.
    """
    poly_set = ONPolynomialSet(ref_el, degree, shape=shape, scale=scale, variant="bubble")
    entity_ids = expansions.polynomial_entity_ids(ref_el, degree, continuity="C0")
    sd = ref_el.get_spatial_dimension()
    dim = sd - codim
    if dim == 1:
        # Apply even / odd reordering on edge bubbles
        indices = []
        for entity in entity_ids[dim]:
            ids = entity_ids[dim][entity]
            indices.extend(ids[::2])
            indices.extend(ids[1::2])
    else:
        indices = list(chain(*entity_ids[dim].values()))

    if shape != ():
        ncomp = numpy.prod(shape)
        dimPk = poly_set.get_num_members() // ncomp
        indices = list((numpy.array(indices)[:, None] + dimPk * numpy.arange(ncomp)[None, :]).flat)
    poly_set = poly_set.take(indices)
    return poly_set



from itertools import chain

import numpy

import gem
from gem.optimise import delta_elimination, sum_factorise, traverse_product
from gem.utils import cached_property

from finat.finiteelementbase import FiniteElementBase


class TensorFiniteElement(FiniteElementBase):

    def __init__(self, element, shape, transpose=False):
        # TODO: Update docstring for arbitrary rank!
        r"""A Finite element whose basis functions have the form:

        .. math::

            \boldsymbol\phi_{i \alpha \beta} = \mathbf{e}_{\alpha} \mathbf{e}_{\beta}^{\mathrm{T}}\phi_i

        Where :math:`\{\mathbf{e}_\alpha,\, \alpha=0\ldots\mathrm{shape[0]}\}` and
        :math:`\{\mathbf{e}_\beta,\, \beta=0\ldots\mathrm{shape[1]}\}` are
        the bases for :math:`\mathbb{R}^{\mathrm{shape[0]}}` and
        :math:`\mathbb{R}^{\mathrm{shape[1]}}` respectively; and
        :math:`\{\phi_i\}` is the basis for the corresponding scalar
        finite element space.

        :param element: The scalar finite element.
        :param shape: The geometric shape of the tensor element.
        :param transpose: Changes the DoF ordering from the
                          Firedrake-style XYZ XYZ XYZ XYZ to the
                          FEniCS-style XXXX YYYY ZZZZ.  That is,
                          tensor shape indices come before the scalar
                          basis function indices when transpose=True.

        :math:`\boldsymbol\phi_{i\alpha\beta}` is, of course, tensor-valued. If
        we subscript the vector-value with :math:`\gamma\epsilon` then we can write:

        .. math::
           \boldsymbol\phi_{\gamma\epsilon(i\alpha\beta)} = \delta_{\gamma\alpha}\delta_{\epsilon\beta}\phi_i

        This form enables the simplification of the loop nests which
        will eventually be created, so it is the form we employ here."""
        super(TensorFiniteElement, self).__init__()
        self._base_element = element
        self._shape = shape
        self._transpose = transpose

    @property
    def base_element(self):
        """The base element of this tensor element."""
        return self._base_element

    @property
    def cell(self):
        return self._base_element.cell

    @property
    def complex(self):
        return self._base_element.complex

    @property
    def degree(self):
        return self._base_element.degree

    @property
    def formdegree(self):
        return self._base_element.formdegree

    @cached_property
    def _entity_dofs(self):
        dofs = {}
        base_dofs = self._base_element.entity_dofs()
        ndof = int(numpy.prod(self._shape, dtype=int))

        def expand(dofs):
            dofs = tuple(dofs)
            if self._transpose:
                space_dim = self._base_element.space_dimension()
                # Components stride by space dimension of base element
                iterable = ((v + i*space_dim for v in dofs)
                            for i in range(ndof))
            else:
                # Components packed together
                iterable = (range(v*ndof, (v+1)*ndof) for v in dofs)
            yield from chain.from_iterable(iterable)

        for dim in self.cell.get_topology().keys():
            dofs[dim] = dict((k, list(expand(d)))
                             for k, d in base_dofs[dim].items())
        return dofs

    def entity_dofs(self):
        return self._entity_dofs

    def space_dimension(self):
        return int(numpy.prod(self.index_shape))

    @property
    def index_shape(self):
        if self._transpose:
            return self._shape + self._base_element.index_shape
        else:
            return self._base_element.index_shape + self._shape

    @property
    def value_shape(self):
        return self._shape + self._base_element.value_shape

    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        r"""Produce the recipe for basis function evaluation at a set of points :math:`q`:

        .. math::
            \boldsymbol\phi_{(\gamma \epsilon) (i \alpha \beta) q} = \delta_{\alpha \gamma} \delta_{\beta \epsilon}\phi_{i q}

            \nabla\boldsymbol\phi_{(\epsilon \gamma \zeta) (i \alpha \beta) q} = \delta_{\alpha \epsilon} \delta_{\beta \gamma}\nabla\phi_{\zeta i q}
        """
        scalar_evaluation = self._base_element.basis_evaluation
        return self._tensorise(scalar_evaluation(order, ps, entity, coordinate_mapping=coordinate_mapping))

    def point_evaluation(self, order, point, entity=None, coordinate_mapping=None):
        scalar_evaluation = self._base_element.point_evaluation
        return self._tensorise(scalar_evaluation(order, point, entity, coordinate_mapping))

    def _tensorise(self, scalar_evaluation):
        # Old basis function and value indices
        scalar_i = self._base_element.get_indices()
        scalar_vi = self._base_element.get_value_indices()

        # New basis function and value indices
        tensor_i = tuple(gem.Index(extent=d) for d in self._shape)
        tensor_vi = tuple(gem.Index(extent=d) for d in self._shape)

        # Couple new basis function and value indices
        deltas = gem.Delta(tensor_i, tensor_vi)

        if self._transpose:
            index_ordering = tensor_i + scalar_i + tensor_vi + scalar_vi
        else:
            index_ordering = scalar_i + tensor_i + tensor_vi + scalar_vi

        result = {}
        for alpha, expr in scalar_evaluation.items():
            result[alpha] = gem.ComponentTensor(
                gem.Product(deltas, gem.Indexed(expr, scalar_i + scalar_vi)),
                index_ordering
            )
        return result

    @property
    def dual_basis(self):
        base = self.base_element
        Q, points = base.dual_basis

        # Suppose the tensor element has shape (2, 4)
        # These identity matrices may have difference sizes depending the shapes
        # tQ = Q ⊗ 𝟙₂ ⊗ 𝟙₄
        scalar_i = self.base_element.get_indices()
        scalar_vi = self.base_element.get_value_indices()
        tensor_i = tuple(gem.Index(extent=d) for d in self._shape)
        tensor_vi = tuple(gem.Index(extent=d) for d in self._shape)
        # Couple new basis function and value indices
        deltas = gem.Delta(tensor_i, tensor_vi)
        if self._transpose:
            index_ordering = tensor_i + scalar_i + tensor_vi + scalar_vi
        else:
            index_ordering = scalar_i + tensor_i + tensor_vi + scalar_vi

        Qi = Q[scalar_i + scalar_vi]
        tQ = gem.ComponentTensor(Qi*deltas, index_ordering)
        return tQ, points

    def dual_evaluation(self, fn):
        tQ, x = self.dual_basis
        expr = fn(x)
        # Apply targeted sum factorisation and delta elimination to
        # the expression
        sum_indices, factors = delta_elimination(*traverse_product(expr))
        expr = sum_factorise(sum_indices, factors)
        # NOTE: any shape indices in the expression are because the
        # expression is tensor valued.
        assert expr.shape == self.value_shape

        scalar_i = self.base_element.get_indices()
        scalar_vi = self.base_element.get_value_indices()
        tensor_i = tuple(gem.Index(extent=d) for d in self._shape)
        tensor_vi = tuple(gem.Index(extent=d) for d in self._shape)

        if self._transpose:
            index_ordering = tensor_i + scalar_i + tensor_vi + scalar_vi
        else:
            index_ordering = scalar_i + tensor_i + tensor_vi + scalar_vi

        tQi = tQ[index_ordering]
        expri = expr[tensor_i + scalar_vi]
        evaluation = gem.IndexSum(tQi * expri, x.indices + scalar_vi + tensor_i)
        # This doesn't work perfectly, the resulting code doesn't have
        # a minimal memory footprint, although the operation count
        # does appear to be minimal.
        evaluation = gem.optimise.contraction(evaluation)
        return evaluation, scalar_i + tensor_vi

    @property
    def mapping(self):
        return self._base_element.mapping



# Copyright (C) 2019 Cyrus Cheng (Imperial College London)
#
# This file is part of FIAT.
#
# FIAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FIAT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FIAT. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by David A. Ham (david.ham@imperial.ac.uk), 2019
# Modified by Thomas Bendall (thomas.bendall@metoffice.gov.uk) 2021

from sympy import symbols, legendre, Array, diff, binomial, lambdify
import numpy as np
from FIAT.dual_set import make_entity_closure_ids
from FIAT.finite_element import FiniteElement
from FIAT.polynomial_set import mis
from FIAT.reference_element import compute_unflattening_map, flatten_reference_cube

x, y = symbols('x y')
variables = (x, y)
leg = legendre


def triangular_number(n):
    return int((n+1)*n/2)


class BrezziDouglasMariniCube(FiniteElement):
    """
    The Brezzi-Douglas-Marini element on quadrilateral cells.

    :arg ref_el: The reference element.
    :arg k: The degree.
    :arg mapping: A string giving the Piola mapping.
                  Either 'contravariant Piola' or 'covariant Piola'.
    """

    def __init__(self, ref_el, degree, mapping):

        # Check that ref_el and degree are appropriate
        if degree < 1:
            raise Exception("BDMc_k elements only valid for k >= 1")

        flat_el = flatten_reference_cube(ref_el)
        dim = flat_el.get_spatial_dimension()
        if dim != 2:
            raise Exception("BDMc_k elements only valid for dimension 2")

        # Collect the IDs of the reference element entities
        flat_topology = flat_el.get_topology()

        entity_ids = {}
        counter = 0

        for top_dim, entities in flat_topology.items():
            entity_ids[top_dim] = {}
            for entity in entities:
                entity_ids[top_dim][entity] = []

        for j in sorted(flat_topology[1]):
            entity_ids[1][j] = list(range(counter, counter + degree + 1))
            counter += degree + 1

        entity_ids[2][0] = list(range(counter, counter + 2*triangular_number(degree - 1)))
        counter += 2*triangular_number(degree - 1)

        entity_closure_ids = make_entity_closure_ids(flat_el, entity_ids)

        # Set up FiniteElement
        super().__init__(ref_el=ref_el, dual=None,
                         order=degree, formdegree=1,
                         mapping=mapping)

        # Store unflattened entity ID dictionaries
        topology = ref_el.get_topology()
        unflattening_map = compute_unflattening_map(topology)
        unflattened_entity_ids = {}
        unflattened_entity_closure_ids = {}

        for dim, entities in sorted(topology.items()):
            unflattened_entity_ids[dim] = {}
            unflattened_entity_closure_ids[dim] = {}
        for dim, entities in sorted(flat_topology.items()):
            for entity in entities:
                unflat_dim, unflat_entity = unflattening_map[(dim, entity)]
                unflattened_entity_ids[unflat_dim][unflat_entity] = entity_ids[dim][entity]
                unflattened_entity_closure_ids[unflat_dim][unflat_entity] = entity_closure_ids[dim][entity]
        self.entity_ids = unflattened_entity_ids
        self.entity_closure_ids = unflattened_entity_closure_ids
        self._degree = degree
        self.flat_el = flat_el

    def degree(self):
        """Return the degree of the polynomial space."""
        return self._degree

    def get_nodal_basis(self):
        raise NotImplementedError("get_nodal_basis not implemented for BDMCE/F elements")

    def get_dual_set(self):
        raise NotImplementedError("get_dual_set is not implemented for BDMCE/F elements")

    def get_coeffs(self):
        raise NotImplementedError("get_coeffs not implemented for BDMCE/F elements")

    def tabulate(self, order, points, entity=None):
        """Return tabulated values of derivatives up to a given order of
        basis functions at given points.

        :arg order: The maximum order of derivative.
        :arg points: An iterable of points.
        :arg entity: Optional (dimension, entity number) pair
                     indicating which topological entity of the
                     reference element to tabulate on.  If ``None``,
                     tabulated values are computed by geometrically
                     approximating which facet the points are on.
        """
        if entity is None:
            entity = (self.ref_el.get_dimension(), 0)

        entity_dim, entity_id = entity
        transform = self.ref_el.get_entity_transform(entity_dim, entity_id)
        points = np.asarray(list(map(transform, points)))
        npoints, pointdim = points.shape

        # Turn analytic basis functions into python functions via lambdify
        basis_callable = {(0, 0): numpy_lambdify(variables, self.basis[(0, 0)],
                                                 modules="numpy")}

        # Dictionary of values of functions
        phivals = {}
        for o in range(order+1):
            alphas = mis(2, o)
            # Collect basis and derivatives of basis
            for alpha in alphas:
                try:
                    callable = basis_callable[alpha]
                except KeyError:
                    diff_basis = diff(self.basis[(0, 0)], *zip(variables, alpha))
                    callable = numpy_lambdify(variables, diff_basis, modules="numpy")
                    self.basis[alpha] = diff_basis

                # tabulate by passing points through all lambdified functions
                # resulting array has shape (len(self.basis), spatial_dim, npoints)
                T = np.array([[[func_component(point) for point in points]
                               for func_component in func] for func in callable])

                phivals[alpha] = T

        return phivals

    def entity_dofs(self):
        """Return the map of topological entities to degrees of
        freedom for the finite element."""
        return self.entity_ids

    def entity_closure_dofs(self):
        """Return the map of topological entities to degrees of
        freedom on the closure of those entities for the finite element."""
        return self.entity_closure_ids

    def value_shape(self):
        """Return the value shape of the finite element functions."""
        return np.shape(self.basis[(0, 0)][0])

    def dmats(self):
        raise NotImplementedError

    def get_num_members(self, arg):
        raise NotImplementedError

    def space_dimension(self):
        """Return the dimension of the finite element space."""
        return int(len(self.basis[(0, 0)])/2)


def bdmce_edge_basis(deg, dx, dy, x_mid, y_mid):
    """Returns the basis functions associated with DoFs on the
    edges of elements for the HCurl Brezz-Douglas-Marini element on
    quadrilateral cells.

    These were introduced by Brezzi, Douglas, Marini (1985)
    "Two families of mixed finite elements for Second Order Elliptic Problems"

    Following, e.g. Brezzi, Douglas, Fortin, Marini (1987)
    "Efficient rectangular mixed finite elements in two and three space variables"
    For rectangle K and degree j:
    BDM_j(K) = [P_j(K)^2 + Span(curl(xy^{j+1}, x^{j+1}y))] x P_{j-1}(K)

    The resulting basis functions all have a curl whose polynomials are
    of degree (j - 1).

    :arg deg: The element degree.
    :arg dx: A tuple of sympy expressions, expanding the interval in the
             x direction. Probably (1-x, x).
    :arg dy: A tuple of sympy expressions, expanding the interval in the
             y direction. Probably (1-y, y).
    :arg x_mid: A sympy expression, probably 2*x-1.
    :arg y_mid: A sympy expression, probably 2*y-1.
    """

    # For some functions, we need to multiply by a coefficient to ensure that
    # the result curl is of the correct degree. The coefficient of the
    # highest-order term of leg(deg, 2x-1), is binomial(2*deg, deg)
    coeff = binomial(2*deg, deg) / ((deg+1)*binomial(2*deg-2, deg-1))

    basis = tuple([(0, -leg(j, y_mid)*dx[0]) for j in range(deg)] +
                  [(coeff*-leg(deg-1, y_mid)*dy[0]*dy[1], -leg(deg, y_mid)*dx[0])] +
                  [(0, -leg(j, y_mid)*dx[1]) for j in range(deg)] +
                  [(coeff*leg(deg-1, y_mid)*dy[0]*dy[1], -leg(deg, y_mid)*dx[1])] +
                  [(-leg(j, x_mid)*dy[0], 0) for j in range(deg)] +
                  [(-leg(deg, x_mid)*dy[0], coeff*-leg(deg-1, x_mid)*dx[0]*dx[1])] +
                  [(-leg(j, x_mid)*dy[1], 0) for j in range(deg)] +
                  [(-leg(deg, x_mid)*dy[1], coeff*leg(deg-1, x_mid)*dx[0]*dx[1])])

    return basis


def bdmce_face_basis(deg, dx, dy, x_mid, y_mid):
    """Returns the basis functions associated with DoFs on the
    faces of elements for the HCurl Brezz-Douglas-Marini element on
    quadrilateral cells.

    These were introduced by Brezzi, Douglas, Marini (1985)
    "Two families of mixed finite elements for Second Order Elliptic Problems"

    Following, e.g. Brezzi, Douglas, Fortin, Marini (1987)
    "Efficient rectangular mixed finite elements in two and three space variables"
    For rectangle K and degree j:
    BDM_j(K) = [P_j(K)^2 + Span(curl(xy^{j+1}, x^{j+1}y))] x P_{j-1}(K)

    The resulting basis functions all have a curl whose polynomials are
    of degree (j - 1).

    :arg deg: The element degree.
    :arg dx: A tuple of sympy expressions, expanding the interval in the
             x direction. Probably (1-x, x).
    :arg dy: A tuple of sympy expressions, expanding the interval in the
             y direction. Probably (1-y, y).
    :arg x_mid: A sympy expression, probably 2*x-1.
    :arg y_mid: A sympy expression, probably 2*y-1.
    """

    basis = []
    for k in range(2, deg+1):
        for j in range(k-1):
            basis += [(0, leg(j, x_mid)*leg(k-2-j, y_mid)*dx[0]*dx[1])]
            basis += [(leg(k-2-j, x_mid)*leg(j, y_mid)*dy[0]*dy[1], 0)]

    return tuple(basis)


class BrezziDouglasMariniCubeEdge(BrezziDouglasMariniCube):
    """
    The Brezzi-Douglas-Marini HCurl element on quadrilateral cells.

    :arg ref_el: The reference element.
    :arg k: The degree.
    """

    def __init__(self, ref_el, degree):

        bdmce_list = construct_bdmce_basis(ref_el, degree)
        self.basis = {(0, 0): Array(bdmce_list)}

        super().__init__(ref_el=ref_el, degree=degree, mapping="covariant piola")


class BrezziDouglasMariniCubeFace(BrezziDouglasMariniCube):
    """
    The Brezzi-Douglas-Marini HDiv element on quadrilateral cells.

    :arg ref_el: The reference element.
    :arg k: The degree.
    """

    def __init__(self, ref_el, degree):

        bdmce_list = construct_bdmce_basis(ref_el, degree)

        # BDMCF functions are rotations of BDMCE functions
        bdmcf_list = [[-a[1], a[0]] for a in bdmce_list]
        self.basis = {(0, 0): Array(bdmcf_list)}

        super().__init__(ref_el=ref_el, degree=degree, mapping="contravariant piola")


def construct_bdmce_basis(ref_el, degree):
    """
    Return the basis functions for a particular BDMCE space as a list.

    :arg ref_el: The reference element.
    :arg k: The degree.
    """

    # Extract the vertices from the reference element
    flat_el = flatten_reference_cube(ref_el)
    verts = flat_el.get_vertices()

    # dx on reference quad is (1-x, x)
    dx = ((verts[-1][0] - x) / (verts[-1][0] - verts[0][0]),
          (x - verts[0][0]) / (verts[-1][0] - verts[0][0]))
    # dy on reference quad is (1-y, y)
    dy = ((verts[-1][1] - y) / (verts[-1][1] - verts[0][1]),
          (y - verts[0][1]) / (verts[-1][1] - verts[0][1]))

    # x_mid and y_mid are (2x-1) and (2y-1)
    x_mid = 2*x-(verts[-1][0] + verts[0][0])
    y_mid = 2*y-(verts[-1][1] + verts[0][1])

    # Compute basis functions for BDMcE
    edge_basis = bdmce_edge_basis(degree, dx, dy, x_mid, y_mid)
    face_basis = bdmce_face_basis(degree, dx, dy, x_mid, y_mid) if degree > 1 else ()

    return edge_basis + face_basis


def numpy_lambdify(X, F, modules="numpy", dummify=False):
    '''Unfortunately, SymPy's own lambdify() doesn't work well with
    NumPy in that simple functions like
        lambda x: 1.0,
    when evaluated with NumPy arrays, return just "1.0" instead of
    an array of 1s with the same shape as x. This function does that.
    '''
    try:
        lambda_x = [numpy_lambdify(X, f, modules=modules, dummify=dummify) for f in F]
    except TypeError:  # 'function' object is not iterable
        # SymPy's lambdify also works on functions that return arrays.
        # However, use it componentwise here so we can add 0*x to each
        # component individually. This is necessary to maintain shapes
        # if evaluated with NumPy arrays.
        lmbd_tmp = lambdify(X, F, modules=modules, dummify=dummify)
        lambda_x = lambda u: lmbd_tmp(*[v for v in u]) + 0 * u[0]

    return lambda_x



# Copyright (C) 2008 Robert C. Kirby (Texas Tech University)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Several functions related to the one-dimensional jacobi polynomials:
Evaluation, evaluation of derivatives, plus computation of the roots
via Newton's method.  These mainly are used in defining the expansion
functions over the simplices and in defining quadrature
rules over each domain."""

import numpy


def eval_jacobi(a, b, n, x):
    """Evaluates the nth jacobi polynomial with weight parameters a,b at a
    point x. Recurrence relations implemented from the pseudocode
    given in Karniadakis and Sherwin, Appendix B"""

    if 0 == n:
        # Get zeros of the right shape
        return 0.0 * x + 1.0
    elif 1 == n:
        return 0.5 * (a - b + (a + b + 2.0) * x)
    else:  # 2 <= n
        apb = a + b
        pn2 = 1.0
        pn1 = 0.5 * (a - b + (apb + 2.0) * x)
        p = 0
        for k in range(2, n + 1):
            a1 = 2.0 * k * (k + apb) * (2.0 * k + apb - 2.0)
            a2 = (2.0 * k + apb - 1.0) * (a * a - b * b)
            a3 = (2.0 * k + apb - 2.0)  \
                * (2.0 * k + apb - 1.0) \
                * (2.0 * k + apb)
            a4 = 2.0 * (k + a - 1.0) * (k + b - 1.0) \
                * (2.0 * k + apb)
            a2 = a2 / a1
            a3 = a3 / a1
            a4 = a4 / a1
            p = (a2 + a3 * x) * pn1 - a4 * pn2
            pn2 = pn1
            pn1 = p
        return p


def eval_jacobi_batch(a, b, n, xs):
    """Evaluates all jacobi polynomials with weights a,b
    up to degree n.  xs is a numpy.array of points.
    Returns a two-dimensional array of tabulations, where the
    rows correspond to the Jacobi polynomials and the
    columns correspond to the points."""
    result = numpy.zeros((n + 1, *xs.shape[:-1]), xs.dtype)
    result[0] = 1.0

    if n > 0:
        xsnew = xs.reshape(result.shape[1:])
        result[1] = 0.5 * (a - b + (a + b + 2.0) * xsnew)

        apb = a + b
        for k in range(2, n + 1):
            a1 = 2.0 * k * (k + apb) * (2.0 * k + apb - 2.0)
            a2 = (2.0 * k + apb - 1.0) * (a * a - b * b)
            a3 = (2.0 * k + apb - 2.0)  \
                * (2.0 * k + apb - 1.0) \
                * (2.0 * k + apb)
            a4 = 2.0 * (k + a - 1.0) * (k + b - 1.0) \
                * (2.0 * k + apb)
            a2 = a2 / a1
            a3 = a3 / a1
            a4 = a4 / a1
            result[k] = (a2 + a3 * xsnew) * result[k-1] \
                - a4 * result[k-2]
    return result


def eval_jacobi_deriv(a, b, n, x):
    """Evaluates the first derivative of P_{n}^{a,b} at a point x."""
    if n == 0:
        return 0.0
    else:
        return 0.5 * (a + b + n + 1) * eval_jacobi(a + 1, b + 1, n - 1, x)


def eval_jacobi_deriv_batch(a, b, n, xs):
    """Evaluates the first derivatives of all jacobi polynomials with
    weights a,b up to degree n.  xs is a numpy.array of points.
    Returns a two-dimensional array of points, where the
    rows correspond to the Jacobi polynomials and the
    columns correspond to the points."""
    results = numpy.zeros((n + 1, len(xs)), xs.dtype)
    if n == 0:
        return results
    else:
        results[1:, :] = eval_jacobi_batch(a + 1, b + 1, n - 1, xs)
    for j in range(1, n + 1):
        results[j, :] *= 0.5 * (a + b + j + 1)
    return results



import FIAT
import finat
import numpy
import pytest
from finat.point_set import PointSet
from finat.restricted import r_to_codim
from gem.interpreter import evaluate


def tabulate(element, ps):
    tabulation, = element.basis_evaluation(0, ps).values()
    result, = evaluate([tabulation])
    # Singleton point
    shape = (int(numpy.prod(element.index_shape)), ) + element.value_shape
    return result.arr.reshape(*shape)


def which_dofs(element, restricted):
    edofs = element.entity_dofs()
    rdofs = restricted.entity_dofs()
    keep_e = []
    keep_r = []
    for k in edofs.keys():
        for e, indices in edofs[k].items():
            if rdofs[k][e]:
                assert len(rdofs[k][e]) == len(indices)
                keep_e.extend(indices)
                keep_r.extend(rdofs[k][e])
    return keep_e, keep_r


@pytest.fixture(params=["vertex", "edge", "facet", "interior"], scope="module")
def restriction(request):
    return request.param


@pytest.fixture(params=["tet", "quad", "prism"], scope="module")
def cell(request):
    if request.param == "tet":
        cell = (FIAT.ufc_simplex(3),)
    elif request.param == "quad":
        interval = FIAT.ufc_simplex(1)
        cell = (interval, interval)
    elif request.param == "prism":
        triangle = FIAT.ufc_simplex(2)
        interval = FIAT.ufc_simplex(1)
        cell = (triangle, interval)
    return cell


@pytest.fixture
def ps(cell):
    dim = sum(e.get_spatial_dimension() for e in cell)
    return PointSet([[1/3, 1/4, 1/5][:dim]])


@pytest.fixture(scope="module")
def scalar_element(cell):
    if len(cell) == 1:
        return finat.Lagrange(cell[0], 4)
    else:
        e1, e2 = cell
        return finat.FlattenedDimensions(
            finat.TensorProductElement([
                finat.GaussLobattoLegendre(e1, 3),
                finat.GaussLobattoLegendre(e2, 3)]
            )
        )


@pytest.fixture(scope="module")
def hdiv_element(cell):
    if len(cell) == 1:
        return finat.RaviartThomas(cell[0], 3, variant="integral(3)")
    else:
        e1, e2 = cell
        element = finat.GaussLobattoLegendre if e1.get_spatial_dimension() == 1 else finat.RaviartThomas
        return finat.FlattenedDimensions(
            finat.EnrichedElement([
                finat.HDivElement(
                    finat.TensorProductElement([
                        element(e1, 3),
                        finat.GaussLegendre(e2, 3)])),
                finat.HDivElement(
                    finat.TensorProductElement([
                        finat.GaussLegendre(e1, 3),
                        finat.GaussLobattoLegendre(e2, 3)]))
            ]))


@pytest.fixture(scope="module")
def hcurl_element(cell):
    if len(cell) == 1:
        return finat.Nedelec(cell[0], 3, variant="integral(3)")
    else:
        e1, e2 = cell
        element = finat.GaussLegendre if e1.get_spatial_dimension() == 1 else finat.Nedelec
        return finat.FlattenedDimensions(
            finat.EnrichedElement([
                finat.HCurlElement(
                    finat.TensorProductElement([
                        finat.GaussLobattoLegendre(e1, 3),
                        finat.GaussLegendre(e2, 3)])),
                finat.HCurlElement(
                    finat.TensorProductElement([
                        element(e1, 3),
                        finat.GaussLobattoLegendre(e2, 3)]))
            ]))


def run_restriction(element, restriction, ps):
    try:
        restricted = finat.RestrictedElement(element, restriction)
    except ValueError:
        # No dofs.
        # Check that the original element had no dofs in all the relevant slots.
        dim = element.cell.get_spatial_dimension()
        lo_codim = r_to_codim(restriction, dim)
        hi_codim = (lo_codim if restriction == "interior" else dim)
        edofs = element.entity_dofs()
        for entity_dim, dof_numbering in edofs.items():
            try:
                entity_codim = dim - sum(entity_dim)
            except TypeError:
                entity_codim = dim - entity_dim
            if lo_codim <= entity_codim <= hi_codim:
                assert all(len(i) == 0 for i in dof_numbering.values())
    else:
        e = tabulate(element, ps)
        r = tabulate(restricted, ps)
        keep_e, keep_r = which_dofs(element, restricted)
        assert numpy.allclose(e[keep_e, ...], r[keep_r, ...])


def test_scalar_restriction(scalar_element, restriction, ps):
    run_restriction(scalar_element, restriction, ps)


def test_hdiv_restriction(hdiv_element, restriction, ps):
    run_restriction(hdiv_element, restriction, ps)


def test_hcurl_restriction(hcurl_element, restriction, ps):
    run_restriction(hcurl_element, restriction, ps)



import FIAT
from math import comb
from gem import ListTensor

from finat.fiat_elements import ScalarFiatElement
from finat.physically_mapped import Citations, identity, PhysicallyMappedElement
from finat.argyris import _vertex_transform, _edge_transform, _normal_tangential_transform
from copy import deepcopy


class HsiehCloughTocher(PhysicallyMappedElement, ScalarFiatElement):
    def __init__(self, cell, degree=3, avg=False):
        if Citations is not None:
            Citations().register("Clough1965")
            if degree > 3:
                Citations().register("Groselj2022")
        self.avg = avg
        super().__init__(FIAT.HsiehCloughTocher(cell, degree))

    def basis_transformation(self, coordinate_mapping):
        V = identity(self.space_dimension())

        sd = self.cell.get_dimension()
        top = self.cell.get_topology()

        vorder = 1
        eorder = self.degree - 3
        voffset = comb(sd + vorder, vorder)
        _vertex_transform(V, vorder, self.cell, coordinate_mapping)
        _edge_transform(V, vorder, eorder, self.cell, coordinate_mapping, avg=self.avg)

        # Patch up conditioning
        h = coordinate_mapping.cell_size()
        for v in sorted(top[0]):
            s = voffset*v + 1
            V[:, s:s+sd] *= 1/h[v]
        return ListTensor(V.T)


class ReducedHsiehCloughTocher(PhysicallyMappedElement, ScalarFiatElement):
    def __init__(self, cell, degree=3):
        if Citations is not None:
            Citations().register("Clough1965")
        super().__init__(FIAT.HsiehCloughTocher(cell, reduced=True))

        reduced_dofs = deepcopy(self._element.entity_dofs())
        sd = cell.get_spatial_dimension()
        for entity in reduced_dofs[sd-1]:
            reduced_dofs[sd-1][entity] = []
        self._entity_dofs = reduced_dofs

    def basis_transformation(self, coordinate_mapping):
        sd = self.cell.get_spatial_dimension()
        top = self.cell.get_topology()
        numbf = self._element.space_dimension()
        ndof = self.space_dimension()
        # rectangular to toss out the constraint dofs
        V = identity(numbf, ndof)

        vorder = 1
        voffset = comb(sd + vorder, vorder)
        _vertex_transform(V, vorder, self.cell, coordinate_mapping)

        # Jacobian at barycenter
        bary, = self.cell.make_points(sd, 0, sd+1)
        J = coordinate_mapping.jacobian_at(bary)
        detJ = coordinate_mapping.detJ_at(bary)
        for e in sorted(top[1]):
            s = len(top[0]) * voffset + e
            v0id, v1id = (v * voffset for v in top[1][e])
            Bnn, Bnt, Jt = _normal_tangential_transform(self.cell, J, detJ, e)

            # vertex points
            V[s, v0id] = 1/5 * Bnt
            V[s, v1id] = -1 * V[s, v0id]

            # vertex derivatives
            for i in range(sd):
                V[s, v1id+1+i] = 1/10 * Bnt * Jt[i]
                V[s, v0id+1+i] = V[s, v1id+1+i]

        # Patch up conditioning
        h = coordinate_mapping.cell_size()
        for v in sorted(top[0]):
            s = voffset * v + 1
            V[:, s:s+sd] *= 1/h[v]
        return ListTensor(V.T)

    # This wipes out the edge dofs.  FIAT gives a 12 DOF element
    # because we need some extra functions to help with transforming
    # under the edge constraint.  However, we only have a 9 DOF
    # element.
    def entity_dofs(self):
        return self._entity_dofs

    def space_dimension(self):
        return 9



import pytest

from FIAT import ufc_cell
from finat.quadrature import make_quadrature


@pytest.mark.parametrize(
    "cell_name",
    ["interval", "triangle", "interval * interval", "triangle * interval"]
)
def test_quadrature_rules_are_hashable(cell_name):
    ref_cell = ufc_cell(cell_name)
    quadrature1 = make_quadrature(ref_cell, 3)
    quadrature2 = make_quadrature(ref_cell, 3)

    assert quadrature1 is not quadrature2
    assert hash(quadrature1) == hash(quadrature2)
    assert repr(quadrature1) == repr(quadrature2)
    assert quadrature1 == quadrature2



"""Implementation of the Arnold-Winther finite elements."""
import FIAT
import numpy
from gem import ListTensor

from finat.fiat_elements import FiatElement
from finat.physically_mapped import Citations, identity, PhysicallyMappedElement
from finat.piola_mapped import adjugate, normal_tangential_edge_transform, normal_tangential_face_transform


def _facet_transform(fiat_cell, facet_moment_degree, coordinate_mapping):
    sd = fiat_cell.get_spatial_dimension()
    top = fiat_cell.get_topology()
    num_facets = len(top[sd-1])
    dimPk_facet = FIAT.expansions.polynomial_dimension(
        fiat_cell.construct_subelement(sd-1), facet_moment_degree)
    dofs_per_facet = sd * dimPk_facet
    ndofs = num_facets * dofs_per_facet
    V = identity(ndofs)

    bary, = fiat_cell.make_points(sd, 0, sd+1)
    J = coordinate_mapping.jacobian_at(bary)
    detJ = coordinate_mapping.detJ_at(bary)
    if sd == 2:
        transform = normal_tangential_edge_transform
    elif sd == 3:
        transform = normal_tangential_face_transform

    for f in range(num_facets):
        rows = transform(fiat_cell, J, detJ, f)
        for i in range(dimPk_facet):
            s = dofs_per_facet*f + i * sd
            V[s+1:s+sd, s:s+sd] = rows
    return V


def _evaluation_transform(fiat_cell, coordinate_mapping):
    sd = fiat_cell.get_spatial_dimension()
    bary, = fiat_cell.make_points(sd, 0, sd+1)
    J = coordinate_mapping.jacobian_at(bary)
    K = adjugate([[J[i, j] for j in range(sd)] for i in range(sd)])

    indices = [(i, j) for i in range(sd) for j in range(i, sd)]
    ncomp = len(indices)
    W = numpy.zeros((ncomp, ncomp), dtype=object)
    for p, (i, j) in enumerate(indices):
        for q, (m, n) in enumerate(indices):
            W[p, q] = 0.5*(K[i, m] * K[j, n] + K[j, m] * K[i, n])
    W[:, [i != j for i, j in indices]] *= 2
    return W


class ArnoldWintherNC(PhysicallyMappedElement, FiatElement):
    def __init__(self, cell, degree=2):
        if Citations is not None:
            Citations().register("Arnold2003")
        super().__init__(FIAT.ArnoldWintherNC(cell, degree))

    def basis_transformation(self, coordinate_mapping):
        """Note, the extra 3 dofs which are removed here
        correspond to the constraints."""
        numbf = self._element.space_dimension()
        ndof = self.space_dimension()
        V = identity(numbf, ndof)

        V[:12, :12] = _facet_transform(self.cell, 1, coordinate_mapping)

        # Note: that the edge DOFs are scaled by edge lengths in FIAT implies
        # that they are already have the necessary rescaling to improve
        # conditioning.

        return ListTensor(V.T)

    def entity_dofs(self):
        return {0: {0: [],
                    1: [],
                    2: []},
                1: {0: [0, 1, 2, 3], 1: [4, 5, 6, 7], 2: [8, 9, 10, 11]},
                2: {0: [12, 13, 14]}}

    def space_dimension(self):
        return 15


class ArnoldWinther(PhysicallyMappedElement, FiatElement):
    def __init__(self, cell, degree=3):
        if Citations is not None:
            Citations().register("Arnold2002")
        super().__init__(FIAT.ArnoldWinther(cell, degree))

    def basis_transformation(self, coordinate_mapping):
        # The extra 6 dofs removed here correspond to the constraints
        numbf = self._element.space_dimension()
        ndof = self.space_dimension()
        V = identity(numbf, ndof)

        sd = self.cell.get_spatial_dimension()
        W = _evaluation_transform(self.cell, coordinate_mapping)
        ncomp = W.shape[0]

        # Put into the right rows and columns.
        V[0:3, 0:3] = V[3:6, 3:6] = V[6:9, 6:9] = W
        num_verts = sd + 1
        cur = num_verts * ncomp

        Vsub = _facet_transform(self.cell, 1, coordinate_mapping)
        fdofs = Vsub.shape[0]
        V[cur:cur+fdofs, cur:cur+fdofs] = Vsub
        cur += fdofs

        # RESCALING FOR CONDITIONING
        h = coordinate_mapping.cell_size()
        for e in range(num_verts):
            V[:, ncomp*e:ncomp*(e+1)] *= 1/(h[e] * h[e])

        # Note: that the edge DOFs are scaled by edge lengths in FIAT implies
        # that they are already have the necessary rescaling to improve
        # conditioning.
        return ListTensor(V.T)

    def entity_dofs(self):
        return {0: {0: [0, 1, 2],
                    1: [3, 4, 5],
                    2: [6, 7, 8]},
                1: {0: [9, 10, 11, 12], 1: [13, 14, 15, 16], 2: [17, 18, 19, 20]},
                2: {0: [21, 22, 23]}}

    def space_dimension(self):
        return 24



"""Impero is a helper AST for generating C code (or equivalent,
e.g. COFFEE) from GEM.  An Impero expression is a proper tree, not
directed acyclic graph (DAG).  Impero is a helper AST, not a
standalone language; it is incomplete without GEM as its terminals
refer to nodes from GEM expressions.

Trivia:
 - Impero helps translating GEM into an imperative language.
 - Byzantine units in Age of Empires II sometimes say 'Impero?'
   (Command?) after clicking on them.
"""

from abc import ABCMeta, abstractmethod

from gem.node import Node as NodeBase


class Node(NodeBase):
    """Base class of all Impero nodes"""

    __slots__ = ()


class Terminal(Node, metaclass=ABCMeta):
    """Abstract class for terminal Impero nodes"""

    __slots__ = ()

    children = ()

    @abstractmethod
    def loop_shape(self, free_indices):
        """Gives the loop shape, an ordering of indices for an Impero
        terminal.

        :arg free_indices: a callable mapping of GEM expressions to
                           ordered free indices.
        """
        pass


class Evaluate(Terminal):
    """Assign the value of a GEM expression to a temporary."""

    __slots__ = ('expression',)
    __front__ = ('expression',)

    def __init__(self, expression):
        self.expression = expression

    def loop_shape(self, free_indices):
        return free_indices(self.expression)


class Initialise(Terminal):
    """Initialise an :class:`gem.IndexSum`."""

    __slots__ = ('indexsum',)
    __front__ = ('indexsum',)

    def __init__(self, indexsum):
        self.indexsum = indexsum

    def loop_shape(self, free_indices):
        return free_indices(self.indexsum)


class Accumulate(Terminal):
    """Accumulate terms into an :class:`gem.IndexSum`."""

    __slots__ = ('indexsum',)
    __front__ = ('indexsum',)

    def __init__(self, indexsum):
        self.indexsum = indexsum

    def loop_shape(self, free_indices):
        return free_indices(self.indexsum.children[0])


class Noop(Terminal):
    """No-op terminal. Does not generate code, but wraps a GEM
    expression to have a loop shape, thus affects loop fusion."""

    __slots__ = ('expression',)
    __front__ = ('expression',)

    def __init__(self, expression):
        self.expression = expression

    def loop_shape(self, free_indices):
        return free_indices(self.expression)


class Return(Terminal):
    """Save value of GEM expression into an lvalue. Used to "return"
    values from a kernel."""

    __slots__ = ('variable', 'expression')
    __front__ = ('variable', 'expression')

    def __init__(self, variable, expression):
        assert set(variable.free_indices) >= set(expression.free_indices)

        self.variable = variable
        self.expression = expression

    def loop_shape(self, free_indices):
        return free_indices(self.variable)


class ReturnAccumulate(Terminal):
    """Accumulate an :class:`gem.IndexSum` directly into a return
    variable."""

    __slots__ = ('variable', 'indexsum')
    __front__ = ('variable', 'indexsum')

    def __init__(self, variable, indexsum):
        assert set(variable.free_indices) == set(indexsum.free_indices)

        self.variable = variable
        self.indexsum = indexsum

    def loop_shape(self, free_indices):
        return free_indices(self.indexsum.children[0])


class Block(Node):
    """An ordered set of Impero expressions.  Corresponds to a curly
    braces block in C."""

    __slots__ = ('children',)

    def __init__(self, statements):
        self.children = tuple(statements)


class For(Node):
    """For loop with an index which stores its extent, and a loop body
    expression which is usually a :class:`Block`."""

    __slots__ = ('index', 'children')
    __front__ = ('index',)

    def __new__(cls, index, statement):
        # In case of an empty loop, create a Noop instead.
        # Related: https://github.com/coneoproject/COFFEE/issues/98
        assert isinstance(statement, Block)
        if not statement.children:
            # This "works" because the loop_shape of this node is not
            # asked any more.
            return Noop(None)
        else:
            return super(For, cls).__new__(cls)

    def __init__(self, index, statement):
        self.index = index
        self.children = (statement,)



# Copyright (C) 2008 Robert C. Kirby (Texas Tech University)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Principal orthogonal expansion functions as defined by Karniadakis
and Sherwin.  These are parametrized over a reference element so as
to allow users to get coordinates that they want."""

import numpy
import math
from FIAT import reference_element, jacobi


def morton_index2(p, q=0):
    return (p + q) * (p + q + 1) // 2 + q


def morton_index3(p, q=0, r=0):
    return (p + q + r)*(p + q + r + 1)*(p + q + r + 2)//6 + (q + r)*(q + r + 1)//2 + r


def jrc(a, b, n):
    """Jacobi recurrence coefficients"""
    an = (2*n+1+a+b)*(2*n+2+a+b) / (2*(n+1)*(n+1+a+b))
    bn = (a+b)*(a-b)*(2*n+1+a+b) / (2*(n+1)*(n+1+a+b)*(2*n+a+b))
    cn = (n+a)*(n+b)*(2*n+2+a+b) / ((n+1)*(n+1+a+b)*(2*n+a+b))
    return an, bn, cn


def integrated_jrc(a, b, n):
    """Integrated Jacobi recurrence coefficients"""
    if n == 1:
        an = (a + b + 2) / 2
        bn = (a - 3*b - 2) / 2
        cn = 0.0
    else:
        an, bn, cn = jrc(a-1, b+1, n-1)
    return an, bn, cn


def pad_coordinates(ref_pts, embedded_dim):
    """Pad reference coordinates by appending -1.0."""
    return tuple(ref_pts) + (-1.0, )*(embedded_dim - len(ref_pts))


def pad_jacobian(A, embedded_dim):
    """Pad coordinate mapping Jacobian by appending zero rows."""
    A = numpy.pad(A, [(0, embedded_dim - A.shape[0]), (0, 0)])
    return tuple(row[..., None] for row in A)


def jacobi_factors(x, y, z, dx, dy, dz):
    fb = 0.5 * (y + z)
    fa = x + (fb + 1.0)
    fc = fb ** 2
    dfa = dfb = dfc = None
    if dx is not None:
        dfb = 0.5 * (dy + dz)
        dfa = dx + dfb
        dfc = 2 * fb * dfb
    return fa, fb, fc, dfa, dfb, dfc


def dubiner_recurrence(dim, n, order, ref_pts, Jinv, scale, variant=None):
    """Tabulate a Dubiner expansion set using the recurrence from (Kirby 2010).

    :arg dim: The spatial dimension of the simplex.
    :arg n: The polynomial degree.
    :arg order: The maximum order of differentiation.
    :arg ref_pts: An ``ndarray`` with the coordinates on the default (-1, 1)^d simplex.
    :arg Jinv: The inverse of the Jacobian of the coordinate mapping from the default simplex.
    :arg scale: A scale factor that sets the first member of expansion set.
    :arg variant: Choose between the default (None) orthogonal basis,
                  'bubble' for integrated Jacobi polynomials,
                  or 'dual' for the L2-duals of the integrated Jacobi polynomials.

    :returns: A tuple with tabulations of the expansion set and its derivatives.
    """
    if order > 2:
        raise ValueError("Higher order derivatives not supported")
    if variant not in [None, "bubble", "dual"]:
        raise ValueError(f"Invalid variant {variant}")

    if variant == "bubble":
        scale = -scale

    num_members = math.comb(n + dim, dim)
    results = tuple([None] * num_members for i in range(order+1))
    phi, dphi, ddphi = results + (None,) * (2-order)

    outer = lambda x, y: x[:, None, ...] * y[None, ...]
    sym_outer = lambda x, y: outer(x, y) + outer(y, x)

    pad_dim = dim + 2
    dX = pad_jacobian(Jinv, pad_dim)
    phi[0] = sum((ref_pts[i] - ref_pts[i] for i in range(dim)), scale)
    if dphi is not None:
        dphi[0] = (phi[0] - phi[0]) * dX[0]
    if ddphi is not None:
        ddphi[0] = outer(dphi[0], dX[0])
    if dim == 0 or n == 0:
        return results
    if dim > 3 or dim < 0:
        raise ValueError("Invalid number of spatial dimensions")

    beta = 1 if variant == "dual" else 0
    coefficients = integrated_jrc if variant == "bubble" else jrc
    X = pad_coordinates(ref_pts, pad_dim)
    idx = (lambda p: p, morton_index2, morton_index3)[dim-1]
    for codim in range(dim):
        # Extend the basis from codim to codim + 1
        fa, fb, fc, dfa, dfb, dfc = jacobi_factors(*X[codim:codim+3], *dX[codim:codim+3])
        ddfc = 2 * outer(dfb, dfb)
        for sub_index in reference_element.lattice_iter(0, n, codim):
            # handle i = 1
            icur = idx(*sub_index, 0)
            inext = idx(*sub_index, 1)

            if variant == "bubble":
                alpha = 2 * sum(sub_index)
                a = b = -0.5
            else:
                alpha = 2 * sum(sub_index) + len(sub_index)
                if variant == "dual":
                    alpha += 1 + len(sub_index)
                a = 0.5 * (alpha + beta) + 1.0
                b = 0.5 * (alpha - beta)

            factor = a * fa - b * fb
            phi[inext] = factor * phi[icur]
            if dphi is not None:
                dfactor = a * dfa - b * dfb
                dphi[inext] = factor * dphi[icur] + phi[icur] * dfactor
                if ddphi is not None:
                    ddphi[inext] = factor * ddphi[icur] + sym_outer(dphi[icur], dfactor)

            # general i by recurrence
            for i in range(1, n - sum(sub_index)):
                iprev, icur, inext = icur, inext, idx(*sub_index, i + 1)
                a, b, c = coefficients(alpha, beta, i)
                factor = a * fa - b * fb
                phi[inext] = factor * phi[icur] - c * (fc * phi[iprev])
                if dphi is None:
                    continue
                dfactor = a * dfa - b * dfb
                dphi[inext] = (factor * dphi[icur] + phi[icur] * dfactor -
                               c * (fc * dphi[iprev] + phi[iprev] * dfc))
                if ddphi is None:
                    continue
                ddphi[inext] = (factor * ddphi[icur] + sym_outer(dphi[icur], dfactor) -
                                c * (fc * ddphi[iprev] + sym_outer(dphi[iprev], dfc) + phi[iprev] * ddfc))

        # normalize
        d = codim + 1
        shift = 1 if variant == "dual" else 0
        for index in reference_element.lattice_iter(0, n+1, d):
            icur = idx(*index)
            if variant is not None:
                p = index[-1] + shift
                alpha = 2 * (sum(index[:-1]) + d * shift) - 1
                norm2 = (0.5 + d) / d
                if p > 0 and p + alpha > 0:
                    norm2 *= (p + alpha) * (2*p + alpha) / p
            else:
                norm2 = (2*sum(index) + d) / d
            scale = math.sqrt(norm2)
            for result in results:
                result[icur] *= scale
    return results


def C0_basis(dim, n, tabulations):
    """Modify a tabulation of a hierarchical basis to enforce C0-continuity.

    :arg dim: The spatial dimension of the simplex.
    :arg n: The polynomial degree.
    :arg tabulations: An iterable tabulations of the hierarchical basis.

    :returns: A tuple of tabulations of the C0 basis.
    """
    idx = (lambda p: p, morton_index2, morton_index3)[dim-1]
    # Recover facet bubbles
    for phi in tabulations:
        icur = 0
        phi[icur] *= -1
        for inext in range(1, dim+1):
            phi[icur] -= phi[inext]
        if dim == 2:
            for i in range(2, n+1):
                phi[idx(0, i)] -= phi[idx(1, i-1)]
        elif dim == 3:
            for i in range(2, n+1):
                for j in range(0, n+1-i):
                    phi[idx(0, i, j)] -= phi[idx(1, i-1, j)]
                icur = idx(0, 0, i)
                phi[icur] -= phi[idx(0, 1, i-1)]
                phi[icur] -= phi[idx(1, 0, i-1)]

    # Reorder by dimension and entity on the reference simplex
    dofs = list(range(dim+1))
    if dim == 1:
        dofs.extend(range(2, n+1))
    elif dim == 2:
        dofs.extend(idx(1, i-1) for i in range(2, n+1))
        dofs.extend(idx(0, i) for i in range(2, n+1))
        dofs.extend(idx(i, 0) for i in range(2, n+1))

        dofs.extend(idx(i, j) for j in range(1, n+1) for i in range(2, n-j+1))
    else:
        dofs.extend(idx(0, 1, i-1) for i in range(2, n+1))
        dofs.extend(idx(1, 0, i-1) for i in range(2, n+1))
        dofs.extend(idx(1, i-1, 0) for i in range(2, n+1))
        dofs.extend(idx(0, 0, i) for i in range(2, n+1))
        dofs.extend(idx(0, i, 0) for i in range(2, n+1))
        dofs.extend(idx(i, 0, 0) for i in range(2, n+1))

        dofs.extend(idx(1, i-1, j) for j in range(1, n+1) for i in range(2, n-j+1))
        dofs.extend(idx(0, i, j) for j in range(1, n+1) for i in range(2, n-j+1))
        dofs.extend(idx(i, 0, j) for j in range(1, n+1) for i in range(2, n-j+1))
        dofs.extend(idx(i, j, 0) for j in range(1, n+1) for i in range(2, n-j+1))

        dofs.extend(idx(i, j, k) for k in range(1, n+1) for j in range(1, n-k+1) for i in range(2, n-j-k+1))

    return tuple([phi[i] for i in dofs] for phi in tabulations)


def xi_triangle(eta):
    """Maps from [-1,1]^2 to the (-1,1) reference triangle."""
    eta1, eta2 = eta
    xi1 = 0.5 * (1.0 + eta1) * (1.0 - eta2) - 1.0
    xi2 = eta2
    return (xi1, xi2)


def xi_tetrahedron(eta):
    """Maps from [-1,1]^3 to the -1/1 reference tetrahedron."""
    eta1, eta2, eta3 = eta
    xi1 = 0.25 * (1. + eta1) * (1. - eta2) * (1. - eta3) - 1.
    xi2 = 0.5 * (1. + eta2) * (1. - eta3) - 1.
    xi3 = eta3
    return xi1, xi2, xi3


class ExpansionSet(object):
    def __new__(cls, *args, **kwargs):
        """Returns an ExpansionSet instance appropriate for the given
        reference element."""
        if cls is not ExpansionSet:
            return super().__new__(cls)
        try:
            ref_el = args[0]
            expansion_set = {
                reference_element.POINT: PointExpansionSet,
                reference_element.LINE: LineExpansionSet,
                reference_element.TRIANGLE: TriangleExpansionSet,
                reference_element.TETRAHEDRON: TetrahedronExpansionSet,
            }[ref_el.get_shape()]
            return expansion_set(*args, **kwargs)
        except KeyError:
            raise ValueError("Invalid reference element type.")

    def __init__(self, ref_el, scale=None, variant=None):
        self.ref_el = ref_el
        self.variant = variant
        sd = ref_el.get_spatial_dimension()
        top = ref_el.get_topology()
        base_ref_el = reference_element.default_simplex(sd)
        base_verts = base_ref_el.get_vertices()

        self.affine_mappings = [reference_element.make_affine_mapping(
                                ref_el.get_vertices_of_subcomplex(top[sd][cell]),
                                base_verts) for cell in top[sd]]
        if scale is None:
            scale = math.sqrt(1.0 / base_ref_el.volume())
        self.scale = scale
        self.variant = variant
        self.continuity = "C0" if variant == "bubble" else None
        self.recurrence_order = 2
        self._dmats_cache = {}
        self._cell_node_map_cache = {}

    def reconstruct(self, ref_el=None, scale=None, variant=None):
        """Reconstructs this ExpansionSet with modified arguments."""
        return ExpansionSet(ref_el or self.ref_el,
                            scale=scale or self.scale,
                            variant=variant or self.variant)

    def get_scale(self, n, cell=0):
        scale = self.scale
        sd = self.ref_el.get_spatial_dimension()
        if isinstance(scale, str):
            vol = self.ref_el.volume_of_subcomplex(sd, cell)
            scale = scale.lower()
            if scale == "orthonormal":
                scale = math.sqrt(1.0 / vol)
            elif scale == "l2 piola":
                scale = 1.0 / vol
        elif n == 0 and sd > 1 and len(self.affine_mappings) == 1:
            # return 1 for n=0 to make regression tests pass
            scale = 1
        return scale

    def get_num_members(self, n):
        return polynomial_dimension(self.ref_el, n, self.continuity)

    def get_cell_node_map(self, n):
        try:
            return self._cell_node_map_cache[n]
        except KeyError:
            cell_node_map = polynomial_cell_node_map(self.ref_el, n, self.continuity)
            return self._cell_node_map_cache.setdefault(n, cell_node_map)

    def _tabulate_on_cell(self, n, pts, order=0, cell=0, direction=None):
        """Returns a dict of tabulations such that
        tabulations[alpha][i, j] = D^alpha phi_i(pts[j])."""
        from FIAT.polynomial_set import mis
        lorder = min(order, self.recurrence_order)
        A, b = self.affine_mappings[cell]
        ref_pts = numpy.add(numpy.dot(pts, A.T), b).T
        Jinv = A if direction is None else numpy.dot(A, direction)[:, None]
        sd = self.ref_el.get_spatial_dimension()
        scale = self.get_scale(n, cell=cell)
        phi = dubiner_recurrence(sd, n, lorder, ref_pts, Jinv,
                                 scale, variant=self.variant)
        if self.continuity == "C0":
            phi = C0_basis(sd, n, phi)

        # Pack linearly independent components into a dictionary
        result = {(0,) * sd: numpy.asarray(phi[0])}
        for r in range(1, len(phi)):
            vr = numpy.transpose(phi[r], tuple(range(1, r+1)) + (0, r+1))
            for indices in numpy.ndindex(vr.shape[:r]):
                alpha = tuple(map(indices.count, range(sd)))
                if alpha not in result:
                    result[alpha] = vr[indices]

        def distance(alpha, beta):
            return sum(ai != bi for ai, bi in zip(alpha, beta))

        # Only use dmats if tabulate failed
        for i in range(len(phi), order + 1):
            dmats = self.get_dmats(n, cell=cell)
            for alpha in mis(sd, i):
                base_alpha = next(a for a in result if sum(a) == i-1 and distance(alpha, a) == 1)
                vals = result[base_alpha]
                for dmat, start, end in zip(dmats, base_alpha, alpha):
                    for j in range(start, end):
                        vals = numpy.dot(dmat.T, vals)
                result[alpha] = vals
        return result

    def _tabulate(self, n, pts, order=0):
        """A version of tabulate() that also works for a single point."""
        pts = numpy.asarray(pts)
        unique = self.continuity is not None and order == 0
        cell_point_map = compute_cell_point_map(self.ref_el, pts, unique=unique)
        phis = {cell: self._tabulate_on_cell(n, pts[ipts], order, cell=cell)
                for cell, ipts in cell_point_map.items()}

        if not self.ref_el.is_macrocell():
            return phis[0]

        if pts.dtype == object:
            # If binning is undefined, scale by the characteristic function of each subcell
            Xi = compute_partition_of_unity(self.ref_el, pts, unique=unique)
            for cell, phi in phis.items():
                for alpha in phi:
                    phi[alpha] *= Xi[cell]
        elif not unique:
            # If binning is not unique, divide by the multiplicity of each point
            mult = numpy.zeros(pts.shape[:-1])
            for cell, ipts in cell_point_map.items():
                mult[ipts] += 1
            for cell, ipts in cell_point_map.items():
                phi = phis[cell]
                for alpha in phi:
                    phi[alpha] /= mult[None, ipts]

        # Insert subcell tabulations into the corresponding submatrices
        idx = lambda *args: args if args[-1] is Ellipsis else numpy.ix_(*args)
        num_phis = self.get_num_members(n)
        cell_node_map = self.get_cell_node_map(n)
        result = {}
        base_phi = tuple(phis.values())[0]
        for alpha in base_phi:
            dtype = base_phi[alpha].dtype
            result[alpha] = numpy.zeros((num_phis, *pts.shape[:-1]), dtype=dtype)
            for cell in cell_point_map:
                ibfs = cell_node_map[cell]
                ipts = cell_point_map[cell]
                result[alpha][idx(ibfs, ipts)] += phis[cell][alpha]
        return result

    def tabulate_normal_jumps(self, n, ref_pts, facet, order=0):
        """Tabulates the normal derivative jumps on reference points on a facet.

        :arg n: the polynomial degree.
        :arg ref_pts: an iterable of points on the reference facet.
        :arg facet: the facet id.
        :kwarg order: the order of differentiation.

        :returns: a numpy array of tabulations of normal derivative jumps.
        """
        sd = self.ref_el.get_spatial_dimension()
        transform = self.ref_el.get_entity_transform(sd-1, facet)
        pts = transform(ref_pts)
        cell_point_map = compute_cell_point_map(self.ref_el, pts, unique=False)
        cell_node_map = self.get_cell_node_map(n)

        num_phis = self.get_num_members(n)
        results = numpy.zeros((order+1, num_phis, *pts.shape[:-1]))

        for cell in cell_point_map:
            ipts = cell_point_map[cell]
            ibfs = cell_node_map[cell]
            normal = self.ref_el.compute_normal(facet, cell=cell)
            side = numpy.dot(normal, self.ref_el.compute_normal(facet))
            phi = self._tabulate_on_cell(n, pts[ipts], order, cell=cell)
            v0 = phi[(0,)*sd]
            for r in range(order+1):
                vr = numpy.zeros((sd,)*r + v0.shape, dtype=v0.dtype)
                for index in numpy.ndindex(vr.shape[:r]):
                    vr[index] = phi[tuple(map(index.count, range(sd)))]
                for _ in range(r):
                    vr = numpy.tensordot(normal, vr, axes=(0, 0))

                indices = numpy.ix_(ibfs, ipts)
                if r % 2 == 0 and side < 0:
                    results[r][indices] -= vr
                else:
                    results[r][indices] += vr
        return results

    def tabulate_jumps(self, n, points, order=0):
        """Tabulates derivative jumps on given points.

        :arg n: the polynomial degree.
        :arg points: an iterable of points on the cell complex.
        :kwarg order: the order of differentiation.

        :returns: a dictionary of tabulations of derivative jumps across interior facets.
        """

        from FIAT.polynomial_set import mis
        sd = self.ref_el.get_spatial_dimension()
        num_members = self.get_num_members(n)
        cell_node_map = self.get_cell_node_map(n)
        cell_point_map = compute_cell_point_map(self.ref_el, points, unique=False)

        num_jumps = 0
        facet_point_map = {}
        for facet in self.ref_el.get_interior_facets(sd-1):
            try:
                cells = self.ref_el.connectivity[(sd-1, sd)][facet]
                ipts = list(set.intersection(*(set(cell_point_map[c]) for c in cells)))
                if ipts != ():
                    facet_point_map[facet] = ipts
                    num_jumps += len(ipts)
            except KeyError:
                pass

        derivs = {cell: self._tabulate_on_cell(n, points, order=order, cell=cell)
                  for cell in cell_point_map}

        jumps = {}
        for r in range(order+1):
            cur = 0
            alphas = mis(sd, r)
            jumps[r] = numpy.zeros((num_members, len(alphas) * num_jumps))
            for facet, ipts in facet_point_map.items():
                c0, c1 = self.ref_el.connectivity[(sd-1, sd)][facet]
                for alpha in alphas:
                    ijump = range(cur, cur + len(ipts))
                    jumps[r][numpy.ix_(cell_node_map[c1], ijump)] += derivs[c1][alpha][:, ipts]
                    jumps[r][numpy.ix_(cell_node_map[c0], ijump)] -= derivs[c0][alpha][:, ipts]
                    cur += len(ipts)
        return jumps

    def get_dmats(self, degree, cell=0):
        """Returns a numpy array with the expansion coefficients dmat[k, j, i]
        of the gradient of each member of the expansion set:
            d/dx_k phi_j = sum_i dmat[k, j, i] phi_i.
        """
        from FIAT.polynomial_set import mis
        key = (degree, cell)
        cache = self._dmats_cache
        try:
            return cache[key]
        except KeyError:
            pass
        if degree == 0:
            return cache.setdefault(key, numpy.zeros((self.ref_el.get_spatial_dimension(), 1, 1), "d"))

        D = self.ref_el.get_dimension()
        top = self.ref_el.get_topology()
        verts = self.ref_el.get_vertices_of_subcomplex(top[D][cell])
        pts = reference_element.make_lattice(verts, degree, variant="gl")
        v = self._tabulate_on_cell(degree, pts, order=1, cell=cell)
        dv = [numpy.transpose(v[alpha]) for alpha in mis(D, 1)]
        dmats = numpy.linalg.solve(numpy.transpose(v[(0,) * D]), dv)
        return cache.setdefault(key, dmats)

    def tabulate(self, n, pts):
        if len(pts) == 0:
            return numpy.array([])
        sd = self.ref_el.get_spatial_dimension()
        return self._tabulate(n, pts)[(0,) * sd]

    def tabulate_derivatives(self, n, pts):
        from FIAT.polynomial_set import mis
        vals = self._tabulate(n, pts, order=1)
        # Create the ordinary data structure.
        sd = self.ref_el.get_spatial_dimension()
        v = vals[(0,) * sd]
        dv = [vals[alpha] for alpha in mis(sd, 1)]
        data = [[(v[i, j], [vi[i, j] for vi in dv])
                 for j in range(v.shape[1])]
                for i in range(v.shape[0])]
        return data

    def tabulate_jet(self, n, pts, order=1):
        vals = self._tabulate(n, pts, order=order)
        # Create the ordinary data structure.
        sd = self.ref_el.get_spatial_dimension()
        v0 = vals[(0,) * sd]
        data = [v0]
        for r in range(1, order+1):
            vr = numpy.zeros((sd,) * r + v0.shape, dtype=v0.dtype)
            for index in numpy.ndindex(vr.shape[:r]):
                vr[index] = vals[tuple(map(index.count, range(sd)))]
            data.append(vr.transpose((r, r+1) + tuple(range(r))))
        return data

    def __eq__(self, other):
        return (type(self) is type(other) and
                self.ref_el == other.ref_el and
                self.continuity == other.continuity)


class PointExpansionSet(ExpansionSet):
    """Evaluates the point basis on a point reference element."""
    def __init__(self, ref_el, **kwargs):
        if ref_el.get_spatial_dimension() != 0:
            raise ValueError("Must have a point")
        super().__init__(ref_el, **kwargs)

    def _tabulate_on_cell(self, n, pts, order=0, cell=0, direction=None):
        """Returns a dict of tabulations such that
        tabulations[alpha][i, j] = D^alpha phi_i(pts[j])."""
        assert n == 0 and order == 0
        return {(): numpy.ones((1, len(pts)))}


class LineExpansionSet(ExpansionSet):
    """Evaluates the Legendre basis on a line reference element."""
    def __init__(self, ref_el, **kwargs):
        if ref_el.get_spatial_dimension() != 1:
            raise Exception("Must have a line")
        super().__init__(ref_el, **kwargs)

    def _tabulate_on_cell(self, n, pts, order=0, cell=0, direction=None):
        """Returns a dict of tabulations such that
        tabulations[alpha][i, j] = D^alpha phi_i(pts[j])."""
        if self.variant is not None:
            return super()._tabulate_on_cell(n, pts, order=order, cell=cell, direction=direction)

        A, b = self.affine_mappings[cell]
        Jinv = A[0, 0] if direction is None else numpy.dot(A, direction)
        xs = numpy.add(numpy.dot(pts, A.T), b)
        results = {}
        scale = self.get_scale(n, cell=cell) * numpy.sqrt(2 * numpy.arange(n+1) + 1)
        for k in range(order+1):
            v = numpy.zeros((n + 1, *xs.shape[:-1]), xs.dtype)
            if n >= k:
                v[k:] = jacobi.eval_jacobi_batch(k, k, n-k, xs)
            for p in range(n + 1):
                v[p] *= scale[p]
                scale[p] *= 0.5 * (p + k + 1) * Jinv
            results[(k,)] = v
        return results


class TriangleExpansionSet(ExpansionSet):
    """Evaluates the orthonormal Dubiner basis on a triangular
    reference element."""
    def __init__(self, ref_el, **kwargs):
        if ref_el.get_spatial_dimension() != 2:
            raise Exception("Must have a triangle")
        super().__init__(ref_el, **kwargs)


class TetrahedronExpansionSet(ExpansionSet):
    """Collapsed orthonormal polynomial expansion on a tetrahedron."""
    def __init__(self, ref_el, **kwargs):
        if ref_el.get_spatial_dimension() != 3:
            raise Exception("Must be a tetrahedron")
        super().__init__(ref_el, **kwargs)


def polynomial_dimension(ref_el, n, continuity=None):
    """Returns the dimension of the space of polynomials of degree no
    greater than n on the reference complex."""
    if ref_el.get_shape() == reference_element.POINT:
        if n > 0:
            raise ValueError("Only degree zero polynomials supported on point elements.")
        return 1
    top = ref_el.get_topology()

    if isinstance(continuity, dict):
        space_dimension = sum(len(continuity[dim][0]) * len(top[dim]) for dim in top)
    elif continuity == "C0":
        space_dimension = sum(math.comb(n - 1, dim) * len(top[dim]) for dim in top)
    else:
        dim = ref_el.get_spatial_dimension()
        space_dimension = math.comb(n + dim, dim) * len(top[dim])
    return space_dimension


def polynomial_entity_ids(ref_el, n, continuity=None):
    """Maps entites of a cell complex to members of a polynomial basis.

    :arg ref_el: a SimplicialComplex.
    :arg n: the polynomial degree of the expansion set.
    :arg continuity: the continuity of the expansion set.
    :returns: a dict of dicts mapping dimension and entity id to basis functions.
    """
    top = ref_el.get_topology()
    sd = ref_el.get_spatial_dimension()
    entity_ids = {}
    cur = 0
    for dim in sorted(top):
        if isinstance(continuity, dict):
            dofs, = set(len(continuity[dim][entity]) for entity in continuity[dim])
        elif continuity == "C0":
            dofs = math.comb(n - 1, dim)
        else:
            # DG numbering
            dofs = math.comb(n + dim, dim) if dim == sd else 0
        entity_ids[dim] = {}
        for entity in sorted(top[dim]):
            entity_ids[dim][entity] = list(range(cur, cur + dofs))
            cur += dofs
    return entity_ids


def polynomial_cell_node_map(ref_el, n, continuity=None):
    """Maps cells on a simplicial complex to members of a polynomial basis.

    :arg ref_el: a SimplicialComplex.
    :arg n: the polynomial degree of the expansion set.
    :arg continuity: the continuity of the expansion set.
    :returns: a numpy array mapping cell id to basis functions supported on that cell.
    """
    top = ref_el.get_topology()
    sd = ref_el.get_spatial_dimension()

    entity_ids = polynomial_entity_ids(ref_el, n, continuity)
    ref_entity_ids = polynomial_entity_ids(ref_el.construct_subelement(sd), n, continuity)

    num_cells = len(top[sd])
    dofs_per_cell = sum(len(ref_entity_ids[dim][entity])
                        for dim in ref_entity_ids for entity in ref_entity_ids[dim])
    cell_node_map = numpy.zeros((num_cells, dofs_per_cell), dtype=int)
    conn = ref_el.get_cell_connectivity()
    for cell in top[sd]:
        for dim in top:
            for ref_entity, entity in enumerate(conn[cell][dim]):
                ref_dofs = ref_entity_ids[dim][ref_entity]
                cell_node_map[cell, ref_dofs] = entity_ids[dim][entity]
    return cell_node_map


def compute_cell_point_map(ref_el, pts, unique=True, tol=1E-12):
    """Maps cells on a simplicial complex to points.
    Points outside the complex are binned to the nearest cell.

    :arg ref_el: a SimplicialComplex.
    :arg pts: an iterable of physical points on the complex.
    :kwarg unique: Are we assigning a unique cell to points on facets?
    :kwarg tol: the absolute tolerance.
    :returns: a dict mapping cell id to the point ids nearest to that cell.
    """
    top = ref_el.get_topology()
    sd = ref_el.get_spatial_dimension()
    if len(top[sd]) == 1:
        return {0: Ellipsis}

    pts = numpy.asarray(pts)
    if pts.dtype == object:
        return {cell: Ellipsis for cell in sorted(top[sd])}

    # The distance to the nearest cell is equal to the distance to the parent cell
    best = ref_el.get_parent().distance_to_point_l1(pts, rescale=True)
    tol = best + tol

    cell_point_map = {}
    for cell in sorted(top[sd]):
        # Bin points based on l1 distance
        pts_near_cell = ref_el.distance_to_point_l1(pts, entity=(sd, cell), rescale=True) < tol
        if len(pts_near_cell.shape) == 0:
            # singleton case
            if pts_near_cell:
                cell_point_map[cell] = Ellipsis
                if unique:
                    break
        else:
            if unique:
                for other in cell_point_map.values():
                    pts_near_cell[other] = False
            ipts = numpy.where(pts_near_cell)[0]
            if len(ipts) > 0:
                cell_point_map[cell] = ipts
    return cell_point_map


def compute_partition_of_unity(ref_el, pt, unique=True, tol=1E-12):
    """Computes the partition of unity functions for each subcell.

    :arg ref_el: a SimplicialComplex.
    :arg pt: a physical point on the complex.
    :kwarg unique: Are we assigning a unique cell to points on facets?
    :kwarg tol: the absolute tolerance.
    :returns: a list of (weighted) characteristic functions for each subcell.
    """
    from sympy import Piecewise
    sd = ref_el.get_spatial_dimension()
    top = ref_el.get_topology()
    # assert singleton point
    pt = pt.reshape((sd,))

    # The distance to the nearest cell is equal to the distance to the parent cell
    best = ref_el.get_parent().distance_to_point_l1(pt, rescale=True)
    tol = best + tol

    # Compute characteristic function of each subcell
    otherwise = []
    masks = []
    for cell in sorted(top[sd]):
        # Bin points based on l1 distance
        pt_near_cell = ref_el.distance_to_point_l1(pt, entity=(sd, cell), rescale=True) < tol
        masks.append(Piecewise(*otherwise, (1.0, pt_near_cell), (0.0, True)))
        if unique:
            otherwise.append((0.0, pt_near_cell))
    # If the point is on a facet, divide the characteristic function by the facet multiplicity
    if not unique:
        mult = sum(masks)
        masks = [m / mult for m in masks]
    return masks



import hashlib
from abc import ABCMeta, abstractmethod
from functools import cached_property

import gem
import numpy
from FIAT.quadrature import GaussLegendreQuadratureLineRule
from FIAT.quadrature_schemes import create_quadrature as fiat_scheme
from FIAT.reference_element import LINE, QUADRILATERAL, TENSORPRODUCT
from gem.utils import safe_repr

from finat.point_set import GaussLegendrePointSet, PointSet, TensorPointSet


def make_quadrature(ref_el, degree, scheme="default"):
    """
    Generate quadrature rule for given reference element
    that will integrate an polynomial of order 'degree' exactly.

    For low-degree (<=6) polynomials on triangles and tetrahedra, this
    uses hard-coded rules, otherwise it falls back to a collapsed
    Gauss scheme on simplices.  On tensor-product cells, it is a
    tensor-product quadrature rule of the subcells.

    :arg ref_el: The FIAT cell to create the quadrature for.
    :arg degree: The degree of polynomial that the rule should
        integrate exactly.
    """
    if ref_el.get_shape() == TENSORPRODUCT:
        try:
            degree = tuple(degree)
        except TypeError:
            degree = (degree,) * len(ref_el.cells)

        assert len(ref_el.cells) == len(degree)
        quad_rules = [make_quadrature(c, d, scheme)
                      for c, d in zip(ref_el.cells, degree)]
        return TensorProductQuadratureRule(quad_rules, ref_el=ref_el)

    if ref_el.get_shape() == QUADRILATERAL:
        return make_quadrature(ref_el.product, degree, scheme)

    if degree < 0:
        raise ValueError("Need positive degree, not %d" % degree)

    if ref_el.get_shape() == LINE and not ref_el.is_macrocell():
        # FIAT uses Gauss-Legendre line quadature, however, since we
        # symbolically label it as such, we wish not to risk attaching
        # the wrong label in case FIAT changes.  So we explicitly ask
        # for Gauss-Legendre line quadature.
        num_points = (degree + 1 + 1) // 2  # exact integration
        fiat_rule = GaussLegendreQuadratureLineRule(ref_el, num_points)
        point_set = GaussLegendrePointSet(fiat_rule.get_points())
        return QuadratureRule(point_set, fiat_rule.get_weights(), ref_el=ref_el, io_ornt_map_tuple=fiat_rule._intrinsic_orientation_permutation_map_tuple)

    fiat_rule = fiat_scheme(ref_el, degree, scheme)
    return QuadratureRule(PointSet(fiat_rule.get_points()), fiat_rule.get_weights(), ref_el=ref_el, io_ornt_map_tuple=fiat_rule._intrinsic_orientation_permutation_map_tuple)


class AbstractQuadratureRule(metaclass=ABCMeta):
    """Abstract class representing a quadrature rule as point set and a
    corresponding set of weights."""

    def __hash__(self):
        return int.from_bytes(hashlib.md5(repr(self).encode()).digest(), byteorder="big")

    def __eq__(self, other):
        return type(other) is type(self) and repr(other) == repr(self)

    @abstractmethod
    def __repr__(self):
        pass

    @property
    @abstractmethod
    def point_set(self):
        """Point set object representing the quadrature points."""

    @property
    @abstractmethod
    def weight_expression(self):
        """GEM expression describing the weights, with the same free indices
        as the point set."""

    @cached_property
    def extrinsic_orientation_permutation_map(self):
        """A map from extrinsic orientations to corresponding axis permutation matrices.

        Notes
        -----
        result[eo] gives the physical axis-reference axis permutation matrix corresponding to
        eo (extrinsic orientation).

        """
        if self.ref_el is None:
            raise ValueError("Must set ref_el")
        return self.ref_el.extrinsic_orientation_permutation_map

    @cached_property
    def intrinsic_orientation_permutation_map_tuple(self):
        """A tuple of maps from intrinsic orientations to corresponding point permutations for each reference cell axis.

        Notes
        -----
        result[axis][io] gives the physical point-reference point permutation array corresponding to
        io (intrinsic orientation) on ``axis``.

        """
        if any(m is None for m in self._intrinsic_orientation_permutation_map_tuple):
            raise ValueError("Must set _intrinsic_orientation_permutation_map_tuple")
        return self._intrinsic_orientation_permutation_map_tuple


class QuadratureRule(AbstractQuadratureRule):
    """Generic quadrature rule with no internal structure."""

    def __init__(self, point_set, weights, ref_el=None, io_ornt_map_tuple=(None, )):
        weights = numpy.asarray(weights)
        assert len(point_set.points) == len(weights)

        self.ref_el = ref_el
        self.point_set = point_set
        self.weights = numpy.asarray(weights)
        self._intrinsic_orientation_permutation_map_tuple = io_ornt_map_tuple

    def __repr__(self):
        return (
            f"{type(self).__name__}("
            f"{self.point_set!r}, "
            f"{safe_repr(self.weights)}, "
            f"{self.ref_el!r}, "
            f"{self._intrinsic_orientation_permutation_map_tuple!r}"
            ")"
        )

    @cached_property
    def point_set(self):
        pass  # set at initialisation

    @cached_property
    def weight_expression(self):
        return gem.Indexed(gem.Literal(self.weights), self.point_set.indices)


class TensorProductQuadratureRule(AbstractQuadratureRule):
    """Quadrature rule which is a tensor product of other rules."""

    def __init__(self, factors, ref_el=None):
        self.ref_el = ref_el
        self.factors = tuple(factors)
        self._intrinsic_orientation_permutation_map_tuple = tuple(
            m
            for factor in factors
            for m in factor._intrinsic_orientation_permutation_map_tuple
        )

    def __repr__(self):
        return f"{type(self).__name__}({self.factors!r}, {self.ref_el!r})"

    @cached_property
    def point_set(self):
        return TensorPointSet(q.point_set for q in self.factors)

    @cached_property
    def weight_expression(self):
        return gem.Product(*(q.weight_expression for q in self.factors))



# Copyright (C) 2025 Imperial College London and others.
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by Pablo D. Brubeck (brubeck@protonmail.com), 2025

import numpy
from FIAT import finite_element, dual_set, functional, quadrature
from FIAT.reference_element import LINE
from FIAT.orientation_utils import make_entity_permutations_simplex
from FIAT.barycentric_interpolation import LagrangePolynomialSet, get_lagrange_points
from FIAT.gauss_lobatto_legendre import GaussLobattoLegendre


class HistopolationDualSet(dual_set.DualSet):
    r"""The dual basis for 1D histopolation elements.

    We define window functions w_j that satisfy

    \int_K w_j v dx = \ell_j(v)   for all v in P_{k}

    where

    \ell_j(v) = 1/h_j \int_{[x_j, x_{j+1}]} v dx

    is the usual histopolation dual basis.

    The DOFs are defined as integral moments against w_j.
    """
    def __init__(self, ref_el, degree):
        entity_ids = {0: {0: [], 1: []},
                      1: {0: list(range(0, degree+1))}}

        fe = GaussLobattoLegendre(ref_el, degree+1)
        points = get_lagrange_points(fe.dual_basis())
        h = numpy.diff(numpy.reshape(points, (-1,)))
        B = numpy.diag(1.0 / h[:-1], k=-1)
        numpy.fill_diagonal(B, -1.0 / h)

        rule = quadrature.GaussLegendreQuadratureLineRule(ref_el, degree+1)
        self.rule = rule

        phi = fe.tabulate(1, rule.get_points())
        wts = rule.get_weights()
        D = phi[(1, )][:-1]
        A = numpy.dot(numpy.multiply(D, wts), D.T)

        C = numpy.linalg.solve(A, B)
        F = numpy.dot(C.T, D)
        nodes = [functional.IntegralMoment(ref_el, rule, f) for f in F]

        entity_permutations = {}
        entity_permutations[0] = {0: {0: []}, 1: {0: []}}
        entity_permutations[1] = {0: make_entity_permutations_simplex(1, degree + 1)}

        super().__init__(nodes, ref_el, entity_ids, entity_permutations)


class Histopolation(finite_element.CiarletElement):
    """1D discontinuous element with integral DOFs on GLL subgrid."""
    def __init__(self, ref_el, degree):
        if ref_el.shape != LINE:
            raise ValueError("Histopolation elements are only defined in one dimension.")

        dual = HistopolationDualSet(ref_el, degree)
        poly_set = LagrangePolynomialSet(ref_el, dual.rule.pts)
        formdegree = ref_el.get_spatial_dimension()  # n-form
        super().__init__(poly_set, dual, degree, formdegree)



import FIAT
from gem import ListTensor

from finat.argyris import _edge_transform
from finat.fiat_elements import ScalarFiatElement
from finat.physically_mapped import Citations, identity, PhysicallyMappedElement


class QuadraticPowellSabin6(PhysicallyMappedElement, ScalarFiatElement):
    def __init__(self, cell, degree=2):
        if Citations is not None:
            Citations().register("PowellSabin1977")
        super().__init__(FIAT.QuadraticPowellSabin6(cell))

    def basis_transformation(self, coordinate_mapping):
        Js = [coordinate_mapping.jacobian_at(vertex)
              for vertex in self.cell.get_vertices()]

        h = coordinate_mapping.cell_size()

        d = self.cell.get_dimension()
        M = identity(self.space_dimension())

        cur = 0
        for i in range(d+1):
            cur += 1  # skip the vertex
            J = Js[i]
            for j in range(d):
                for k in range(d):
                    M[cur+j, cur+k] = J[j, k] / h[i]
            cur += d

        return ListTensor(M)


class QuadraticPowellSabin12(PhysicallyMappedElement, ScalarFiatElement):
    def __init__(self, cell, degree=2, avg=False):
        self.avg = avg
        if Citations is not None:
            Citations().register("PowellSabin1977")
        super().__init__(FIAT.QuadraticPowellSabin12(cell))

    def basis_transformation(self, coordinate_mapping):
        J = coordinate_mapping.jacobian_at([1/3, 1/3])

        V = identity(self.space_dimension())

        sd = self.cell.get_dimension()
        top = self.cell.get_topology()
        voffset = sd + 1
        for v in sorted(top[0]):
            s = voffset * v
            for i in range(sd):
                for j in range(sd):
                    V[s+1+i, s+1+j] = J[j, i]

        _edge_transform(V, 1, 0, self.cell, coordinate_mapping, avg=self.avg)

        # Patch up conditioning
        h = coordinate_mapping.cell_size()
        for v in sorted(top[0]):
            for k in range(sd):
                V[:, voffset*v+1+k] /= h[v]
        return ListTensor(V.T)



from .fiat_elements import (Bernstein, Bubble, BrezziDouglasFortinMarini,                    # noqa: F401
                            BrezziDouglasMarini, BrezziDouglasMariniCubeEdge,                # noqa: F401
                            BrezziDouglasMariniCubeFace, CrouzeixRaviart,                    # noqa: F401
                            DiscontinuousLagrange, DiscontinuousTaylor, DPC,                 # noqa: F401
                            FacetBubble, GopalakrishnanLedererSchoberlFirstKind,             # noqa: F401
                            GopalakrishnanLedererSchoberlSecondKind, Histopolation,          # noqa: F401
                            HDivTrace, HellanHerrmannJohnson, KongMulderVeldhuizen,          # noqa: F401
                            Lagrange, Real, Serendipity,                                     # noqa: F401
                            TrimmedSerendipityCurl, TrimmedSerendipityDiv,                   # noqa: F401
                            TrimmedSerendipityEdge, TrimmedSerendipityFace,                  # noqa: F401
                            Nedelec, NedelecSecondKind, RaviartThomas, Regge)                # noqa: F401

from .argyris import Argyris                                       # noqa: F401
from .aw import ArnoldWinther, ArnoldWintherNC                     # noqa: F401
from .hz import HuZhang                                            # noqa: F401
from .bell import Bell                                             # noqa: F401
from .bernardi_raugel import BernardiRaugel, BernardiRaugelBubble  # noqa: F401
from .hct import HsiehCloughTocher, ReducedHsiehCloughTocher       # noqa: F401
from .arnold_qin import ArnoldQin, ReducedArnoldQin                # noqa: F401
from .christiansen_hu import ChristiansenHu                        # noqa: F401
from .alfeld_sorokina import AlfeldSorokina                        # noqa: F401
from .guzman_neilan import GuzmanNeilanFirstKindH1, GuzmanNeilanSecondKindH1, GuzmanNeilanBubble, GuzmanNeilanH1div  # noqa: F401
from .powell_sabin import QuadraticPowellSabin6, QuadraticPowellSabin12  # noqa: F401
from .hermite import Hermite                                       # noqa: F401
from .johnson_mercier import JohnsonMercier                        # noqa: F401
from .mtw import MardalTaiWinther                                  # noqa: F401
from .morley import Morley                                         # noqa: F401
from .direct_serendipity import DirectSerendipity                  # noqa: F401
from .spectral import GaussLobattoLegendre, GaussLegendre, Legendre, IntegratedLegendre, FDMLagrange, FDMQuadrature, FDMDiscontinuousLagrange, FDMBrokenH1, FDMBrokenL2, FDMHermite  # noqa: F401
from .tensorfiniteelement import TensorFiniteElement               # noqa: F401
from .tensor_product import TensorProductElement                   # noqa: F401
from .cube import FlattenedDimensions                              # noqa: F401
from .discontinuous import DiscontinuousElement                    # noqa: F401
from .enriched import EnrichedElement                              # noqa: F401
from .hdivcurl import HCurlElement, HDivElement                    # noqa: F401
from .mixed import MixedElement                                    # noqa: F401
from .nodal_enriched import NodalEnrichedElement                   # noqa: F401
from .quadrature_element import QuadratureElement, make_quadrature_element  # noqa: F401
from .restricted import RestrictedElement                          # noqa: F401
from .runtime_tabulated import RuntimeTabulated                    # noqa: F401
from . import quadrature                                           # noqa: F401
from . import cell_tools                                           # noqa: F401
from . import element_factory                                      # noqa: F401
